/*****************************************************************************
 *                        Web3d.org Copyright (c) 2001
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// External imports
import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * Describes the appearance of an object.
 *
 * @author Alan Hudson
 * @version $Revision: 1.17 $
 */
public class Appearance extends NodeComponent
{
    /** The material properites */
    private Material material;

    /** The texture properties */
    private TextureUnit[] textureUnits;

    /** Current number of valid textures */
    private int numTextures;

    /** The shader used by the appearance */
    private Shader shader;

    /** Attributes for rendering polygons */
    private PolygonAttributes polyAttr;

    /** Attributes for rendering lines */
//    private LineAttributes lineAttr;

    /** Visibility flag for this object. If not visible, it's not rendered */
    private boolean visible;

    /**
     * The default constructor.
     */
    public Appearance()
    {
        visible = true;
    }

    /**
     * Set the visibility state of any geometry associated with this instance.
     *
     * @param state true to make the object visible
     */
    public void setVisible(boolean state)
    {
        visible = state;
    }

    /**
     * Check to see if this appearance is making the geometry visible or
     * not.
     *
     * @return true when the geometry is visible
     */
    public boolean isVisible()
    {
        return visible;
    }

    /**
     * Set the material to use. Null will clear the material.
     *
     * @param mat The new material
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setMaterial(Material mat)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(material != null)
            material.setLive(false);

        material = mat;

        if(material != null)
        {
            material.setUpdateHandler(updateHandler);
            material.setLive(alive);
        }
    }

    /**
     * Get the current material in use.
     *
     * @return The current material instance or null
     */
    public Material getMaterial()
    {
        return material;
    }

    /**
     * Set the polygon rendering attributes to use. Null will clear the
     * current attributes.
     *
     * @param attr The new attributes or null
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setPolygonAttributes(PolygonAttributes attr)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(polyAttr != null)
            polyAttr.setLive(false);

        polyAttr = attr;

        if(polyAttr != null)
        {
            polyAttr.setUpdateHandler(updateHandler);
            polyAttr.setLive(alive);
        }
    }

    /**
     * Get the current polygon rendering attributes in use.
     *
     * @return The current attributes instance or null
     */
    public PolygonAttributes getPolygonAttributes()
    {
        return polyAttr;
    }

    /**
     * Set the shader to use. Null will clear the current shader.
     *
     * @param s The new shader instance to use or null
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setShader(Shader s)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(shader != null)
            shader.setLive(false);

        shader = s;

        if(shader != null)
        {
            shader.setUpdateHandler(updateHandler);
            shader.setLive(alive);
        }
    }

    /**
     * Get the current shader in use.
     *
     * @return The current shader instance or null
     */
    public Shader getShader()
    {
        return shader;
    }

    /**
     * Set the texture units to use.  Null will disable texturing.
     *
     * @param texUnits The new Texture lists to set
     * @param num The number of valid entries in the array
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setTextureUnits(TextureUnit[] texUnits, int num)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if((textureUnits == null) || (textureUnits.length < num))
            textureUnits = new TextureUnit[num];

        // clear out remaining pointers
        for(int i = numTextures; i < num; i++)
            textureUnits[i] = null;

        for(int i = 0; i < num; i++)
        {
            textureUnits[i] = texUnits[i];
            textureUnits[i].setUpdateHandler(updateHandler);
        }

        numTextures = num;
    }

    /**
     * Get the current number of texture units that are valid
     *
     * @return a positive number
     */
    public int numTextureUnits()
    {
        return numTextures;
    }

    /**
     * Get the texture units used by the material.
     */
    public void getTextureUnits(TextureUnit[] texUnits)
    {
        if(texUnits.length < numTextures)
            throw new IllegalArgumentException("Array size is too small");

        if(numTextures > 0)
            System.arraycopy(textureUnits, 0, texUnits, 0, numTextures);
    }

    /**
     * Set the scenegraph update handler for this node.  It will notify
     * all its children of the value. A null value will clear the current
     * handler.
     *
     * @param handler The instance to use as a handler
     */
    protected void setUpdateHandler(NodeUpdateHandler handler)
    {
        super.setUpdateHandler(handler);

        if(material != null)
            material.setUpdateHandler(handler);

        if(polyAttr != null)
            polyAttr.setUpdateHandler(handler);

        for(int i = 0; i < numTextures; i++)
            textureUnits[i].setUpdateHandler(handler);

        if(shader != null)
            shader.setUpdateHandler(updateHandler);
    }

    /**
     * Issue ogl commands needed for this component.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void renderState(GL gl, GLU glu)
    {
        if(numTextures != 0)
        {

            for(int i = 0; i < numTextures ; i++)
            {
                gl.glPushAttrib(GL.GL_TEXTURE_BIT);
                gl.glActiveTexture(GL.GL_TEXTURE0 + i);
                textureUnits[i].renderState(gl, glu);
            }
        }

        if(polyAttr != null)
            polyAttr.renderState(gl, glu);

        if(material != null)
        {
            gl.glPushAttrib(GL.GL_LIGHTING_BIT);
            material.renderState(gl, glu);
        }

        if(shader != null)
            shader.renderState(gl, glu);
    }

    /**
     * Restore all openGL state.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void restoreState(GL gl, GLU glu)
    {
        if(shader != null)
            shader.restoreState(gl, glu);

        if(material != null)
        {
            material.restoreState(gl, glu);
            gl.glPopAttrib();
        }

        if(polyAttr != null)
            polyAttr.restoreState(gl, glu);

        if(numTextures != 0)
        {
            for(int i = numTextures - 1; i >= 0; i--)
            {
                gl.glActiveTexture(GL.GL_TEXTURE0 + i);
                textureUnits[i].restoreState(gl, glu);
                gl.glPopAttrib();
            }
        }
    }

    /**
     * Notification that this object is live now. Overridden to make sure that
     * the live state of the nodes represents the same state as the parent
     * scene graph.
     *
     * @param state true if this should be marked as live now
     */
    protected void setLive(boolean state)
    {
        if(state)
            liveCount++;
        else if(liveCount > 0)
            liveCount--;

        if((liveCount == 0) || !alive)
        {
            super.setLive(state);

            if(numTextures != 0)
            {
                for(int i = 0; i < numTextures; i++)
                    textureUnits[i].setLive(state);
            }

            if(polyAttr != null)
                polyAttr.setLive(state);

            if(material != null)
                material.setLive(state);

            if(shader != null)
                shader.setLive(state);
        }
    }

    /**
     * Check to see if this node is the same reference as the passed node that
     * is a parent of this node. This is the downwards check to ensure that
     * there is no cyclic scene graph structures at the point where someone
     * adds a node to the scenegraph. When the reference and this are the
     * same, an exception is generated. Since each class may have different
     * lists of child node setups, this should be overriden by any class that
     * can take children, and have the call passed along to the children.
     *
     * @param parent The reference to check against this class
     * @throws CyclicSceneGraphStructureException Equal parent and child
     */
    protected void checkForCyclicChild(SceneGraphObject parent)
        throws CyclicSceneGraphStructureException
    {
        if(parent == this)
            throw new CyclicSceneGraphStructureException();

        for(int i = 0; i < numTextures; i++)
            textureUnits[i].checkForCyclicChild(parent);

        // Don't bother with anything else for now. That's just wasted CPU cycles.
    }
}