/*****************************************************************************
 *                        Web3d.org Copyright (c) 2001
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// Standard imports
import java.awt.image.*;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.IntBuffer;
import java.util.HashMap;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * A Texture component that uses raw byte data to be interpreted by the
 * format IDs passed in.
 *
 * @author Alan Hudson
 * @version $Revision: 1.3 $
 */
public class ByteTextureComponent1D extends TextureComponent1D
{
    /** The image data */
    private byte[][] pixels;

    /**
     * Constructs an image with default values.
     */
    public ByteTextureComponent1D()
    {
        this(0, 0, null);
    }

    /**
     * Constructs an image with default values.
     *
     * @param yUp Change the image aroud the Y axis if needed
     */
    public ByteTextureComponent1D(boolean yUp)
    {
        this(0, 0, null);

        invertY = yUp;
    }

    /**
     * Constructs an Image1D using the specified format, width, height and
     * rendered image. If the srcImage[0] is an instance of BufferedImage,
     * the format passed is ignored and the image directly used.
     *
     * @param format The image format.  RGB, RGBA currently
     * @param width The width of the image
     * @param height The height of the image
     * @param srcPixels The image data
     */
    public ByteTextureComponent1D(int format,
                                  int width,
                                  byte[] srcPixels)
    {
        super(1);

        this.width = width;
        this.format = format;

        int size = srcPixels == null ? 0 : srcPixels.length;
        pixels = new byte[1][size];

        System.arraycopy(srcPixels, 0, pixels[0], 0, size);
    }

    /**
     * Constructs an Image1D using the specified format, width, height and
     * rendered image. If the srcImage[0] is an instance of BufferedImage,
     * the format passed is ignored and the image directly used.
     *
     * @param format The image format.  RGB, RGBA currently
     * @param width The width of the image
     * @param height The height of the image
     * @param srcPixels The image data
     * @param numLevels The number of mip-map levels to generate
     */
    public ByteTextureComponent1D(int format,
                                  int width,
                                  byte[][] srcPixels,
                                  int numLevels)
    {
        super(numLevels);

        this.width = width;
        this.format = format;

        pixels = new byte[numLevels][];

        for(int i = 0; i < numLevels; i++)
        {
            int size = srcPixels[i] == null ? 0 : srcPixels[i].length;
            pixels[i] = new byte[size];
            System.arraycopy(srcPixels[i], 0, pixels[i], 0, size);
        }
    }

    /**
     * Constructs an Image1D using the specified format, width, height and
     * rendered image. If the srcImage[0] is an instance of BufferedImage,
     * the format passed is ignored and the image directly used.
     *
     * @param format The image format.  RGB, RGBA currently
     * @param width The width of the image
     * @param height The height of the image
     * @param srcImage The image data
     * @param yUp Change the image aroud the Y axis if needed
     */
    public ByteTextureComponent1D(int format,
                                  int width,
                                  byte[] srcPixels,
                                  boolean yUp)
    {
        this(format, width, srcPixels);

        invertY = yUp;
    }

    /**
     * Update a sub-section of the image data with the new pixel values. Not
     * implemented yet.
     *
     * @param srcX The starting X offset in the existing image space
     * @param srcY The starting Y offset in the existing image space
     * @param destX The starting X offset in the existing image space
     * @param destY The starting Y offset in the existing image space
     * @param width The width of the section to replace
     * @param height The height of the section to replace
     * @param img The image to take data from
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void updateSubImage(int srcX,
                               int srcY,
                               int destX,
                               int destY,
                               int widht,
                               int lvl,
                               byte[] img)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

// Not implemented yet.
    }

    /**
     * Convenience method to convert a buffered image into a NIO array of the
     * corresponding type. Images typically need to be swapped when doing this
     * by the Y axis is in the opposite direction to the one used by OpenGL.
     *
     * @param level Which image level needs to be converted
     * @return an appropriate array type - either IntBuffer or ByteBuffer
     */
    protected ByteBuffer convertImage(int level)
    {
        ByteBuffer ret_val = null;

        ret_val = ByteBuffer.allocateDirect(pixels[level].length);
        ret_val.order(ByteOrder.nativeOrder());
        ret_val.put(pixels[level]);

        return ret_val;
    }
}