/*****************************************************************************
 *                        Web3d.org Copyright (c) 2001
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// External imports
import java.lang.ref.ReferenceQueue;
import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;

// Local imports
// None

/**
 * Interface representing the output of a render pipeline.
 * <p>
 *
 * The output may be any of the traditional types: pBuffer, screen or memory
 * or any non-traditional type like haptic devices, network streams etc.
 *
 *
 * @author Justin Couch
 * @version $Revision: 1.10 $
 */
public interface DrawableSurface
{
    /**
     * Tell the surface to use the given thread as the rendering control thread.
     * Used when you want to clock the renderer to the rest of the pipeline,
     * rather than letting it sit in it's own loop. Due to
     * bugs in JOGL, we have to start the internal rendering threads after the
     * containing window has become visible. See JOGL issue #54 for more details
     * on this bug.
     *
     * @param th The render control thread instance to use or null
     */
    public void setRenderControlThread(Thread th);

    /**
     * Set the background colour that this surface should be cleared to before
     * the drawing step. Colours range from 0 to 1 in the normal manner.
     *
     * @param r The red component of the background clear colour
     * @param g The green component of the background clear colour
     * @param b The blue component of the background clear colour
     * @param a The alpha component of the background clear colour
     */
    public void setClearColor(float r, float g, float b, float a);

    /**
     * Set the queue that holds references to the textures that are possibly
     * ready for deletion. Poll this queue once per frame and delete any
     * items found on the queue. A value of null will clear the queue.
     *
     * @param queue The instance of the queue to use or null
     */
    public void setTextureDeletionQueue(ReferenceQueue queue);

    /**
     * Update the list of items to be rendered to the current list. Draw them
     * at the next oppourtunity.
     *
     * @param commands The list of drawable surfaces to render
     * @param numValid The number of valid items in the array
     */
    public void setDrawableObjects(RenderInstructions[] commands,
                                   int numValid);

    /**
     * Instruct the surface to draw the collected set of nodes now. The
     * registered view environment is used to draw to this surface. If no
     * view is registered, the surface is cleared and then this call is
     * exited. The drawing surface does not swap the buffers at this point.
     */
    public void draw();

    /**
     * Swap the buffers now if the surface supports multiple buffer drawing.
     * For surfaces that don't support multiple buffers, this does nothing.
     */
    public void swap();

    /**
     * Get the underlying object that this surface is rendered to. If it is a
     * screen display device, the surface can be one of AWT Component or
     * Swing JComponent. An off-screen buffer would be a form of AWT Image etc.
     *
     * @return The drawable surface representation
     */
    public Object getSurfaceObject();

    /**
     * Instruct this surface that you have finished with the resources needed
     * and to dispose all rendering resources.
     */
    public void dispose();

    /**
     * Get the surface to VWorld transformation matrix.
     *
     * @param The matrix.  It must be preallocated.
     */
    public void getSurfaceToVWorld(Matrix4f matrix);

    /**
     * Convert a pixel location to surface coordinates.
     *
     * @param x The X coordinate
     * @param y The Y coordinate
     * @param position The converted position.  It must be preallocated.
     */
    public void getPixelLocationInSurface(float x, float y, Point3f position);

    /**
     * Get the Center Eye position in surface coordinates.
     *
     * @param position The current eye position.  It must be preallocated.
     */
    public void getCenterEyeInSurface(Point3f position);
}