/*****************************************************************************
 *                        Web3d.org Copyright (c) 2003
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// Standard imports
import java.util.HashMap;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * Shader container object for the shaders usable with OpenGL 1.4 with the
 * separate vertex and fragment shader code.
 * <p>
 *
 * @author Justin Couch
 * @version $Revision: 1.1 $
 */
public class GL14Shader extends Shader
{
    /** The vertex shader used by the appearance */
    private VertexShader vertexShader;

    /** The fragment shader used by the appearance */
    private FragmentShader fragShader;


    /**
     * Constructs a Shader with default values.
     */
    public GL14Shader()
    {
    }

    /**
     * Set the vertexShader to use.  null will clear the vertexShader
     *
     * @param shader The new shader instance to use or null
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setVertexShader(VertexShader shader)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(vertexShader != null)
            vertexShader.setLive(false);

        vertexShader = shader;

        if(vertexShader != null)
        {
            vertexShader.setUpdateHandler(updateHandler);
            vertexShader.setLive(alive);
        }
    }

    /**
     * Get the current vertexShader in use.
     *
     * @return The current shader instance or null
     */
    public VertexShader getVertexShader()
    {
        return vertexShader;
    }

    /**
     * Set the fragment shader to use.  null will clear the shader.
     *
     * @param shader The new shader instance to use or null
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setFragmentShader(FragmentShader shader)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(fragShader != null)
            fragShader.setLive(false);

        fragShader = shader;

        if(fragShader != null)
        {
            fragShader.setUpdateHandler(updateHandler);
            fragShader.setLive(alive);
        }
    }

    /**
     * Get the current fragShader in use.
     *
     * @return The current shader instance or null
     */
    public FragmentShader getFragmentShader()
    {
        return fragShader;
    }

    /**
     * Set the scenegraph update handler for this node.  It will notify
     * all its children of the value. A null value will clear the current
     * handler.
     *
     * @param handler The instance to use as a handler
     */
    protected void setUpdateHandler(NodeUpdateHandler handler)
    {
        super.setUpdateHandler(handler);

        if(fragShader != null)
            fragShader.setUpdateHandler(updateHandler);

        if(vertexShader != null)
            vertexShader.setUpdateHandler(updateHandler);
    }

    /**
     * Issue ogl commands needed for this component.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void renderState(GL gl, GLU glu)
    {
        if(vertexShader != null)
            vertexShader.renderState(gl, glu);

        if(fragShader != null)
            fragShader.renderState(gl, glu);
    }

    /**
     * Restore all openGL state.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void restoreState(GL gl, GLU glu)
    {
        if(fragShader != null)
            fragShader.restoreState(gl, glu);

        if(vertexShader != null)
            vertexShader.restoreState(gl, glu);
    }

    /**
     * Notification that this object is live now. Overridden to make sure that
     * the live state of the nodes represents the same state as the parent
     * scene graph.
     *
     * @param state true if this should be marked as live now
     */
    protected void setLive(boolean state)
    {
        if(state)
            liveCount++;
        else if(liveCount > 0)
            liveCount--;

        if((liveCount == 0) || !alive)
        {
            super.setLive(state);

            if(vertexShader != null)
                vertexShader.setLive(state);

            if(fragShader != null)
                fragShader.setLive(state);
        }
    }

}