/*****************************************************************************
 *                        Web3d.org Copyright (c) 2001
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// Standard imports
import java.util.ArrayList;

// Application specific imports

/**
 * Geometry is an abstract class that specifies the geometry component information
 * required by a Shape3D node.
 *
 * @author Alan Hudson
 * @version $Revision: 1.8 $
 */
public abstract class Geometry extends NodeComponent
    implements PickableObject
{
    /** Message when the setParent does not receive a group */
    private static final String NULL_PARENT_MSG =
        "Attempting to set a null parent into a geometry node";

    /** Message when attempting to pick the object at the wrong time */
    private static final String PICK_TIMING_MSG =
        "Picking not permitted right now. Picking is only permitted during " +
        "the update observer callbacks.";

    /**
     * You are attempting to pick a node you have marked as not-pickable, yet
     * called the pick method on this class anyway.
     */
    private static final String PICKABLE_FALSE_MSG =
        "This node has been marked as not pickable by the user";

    /** The initial size of the children list */
    private static final int LIST_START_SIZE = 5;

    /** The increment size of the list if it gets overflowed */
    private static final int LIST_INCREMENT = 5;

    /** Listing of all the parents of this node */
    protected Node[] parentList;

    /** Index to the next place to add items in the nodeList */
    protected int lastParentList;

    /** Update handler for the external code. Not created until needed. */
    protected InternalNodeUpdateListener internalUpdater;

    /** Bounding volume set by the user */
    protected BoundingVolume bounds;

    /** Was the bounds automatically calculated? */
    protected boolean implicitBounds;

    /** Flag indicating if this object is pickable currently */
    protected boolean pickable;

    /**
     * Internal implementation of the InternalNodeUpdateListener. Done as an
     * inner class to hide the calls from public consumption.
     */
    private class InternalUpdater
        implements InternalNodeUpdateListener
    {

        /**
         * Notify this node to update it's bounds now and propogate those new
         * bounds to their parent(s).
         */
        public void updateBoundsAndNotify()
        {
            updateBounds();
        }
    }

    /**
     * The default constructor.
     */
    public Geometry()
    {
        parentList = new Node[LIST_START_SIZE];
        lastParentList = 0;
        implicitBounds = true;
        pickable = true;
    }

    /**
     * Update this node's bounds and then call the parent to update it's
     * bounds. Used to propogate bounds changes from the leaves of the tree
     * to the root. A node implementation may decide when and where to tell
     * the parent(s)s that updates are ready.
     */
    protected abstract void updateBounds();

    /**
     * Internal method to recalculate the implicit bounds of this Node. By
     * default the bounds are a point sphere, so derived classes should
     * override this method with something better.
     */
    protected abstract void recomputeBounds();

    /**
     * Mark this node as having dirty bounds due to it's geometry having
     * changed.
     */
    protected abstract void markBoundsDirty();

    /**
     * Notify the node that you have updates to the node that might alter
     * its bounds.
     *
     * @param l The change requestor
     */
    public void boundsChanged(NodeUpdateListener l)
    {
        // Ignore till live
        if(updateHandler == null)
            return;

        if(internalUpdater == null)
            internalUpdater = new InternalUpdater();

        updateHandler.boundsChanged(l, this, internalUpdater);

        for(int i = 0; i < lastParentList; i++)
            parentList[i].markBoundsDirty();
    }

    /**
     * Set the bounds to the given explicit value. When set, auto computation
     * of the bounds of this node is turned off. A value of null can be used
     * to clear the current explicit bounds and return to auto computation.
     *
     * @param b The new bounds to use or null to clear
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setBounds(BoundingVolume b)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        bounds = b;
        implicitBounds = (bounds == null);
    }

    /**
     * Get the currently set bounds for this object. If no explicit bounds have
     * been set, then an implicit set of bounds is returned based on the
     * current scene graph state.
     *
     * @return The current bounds of this object
     */
    public BoundingVolume getBounds()
    {
        // need to check and set a new instance here if needed.
        if(implicitBounds && bounds == null)
            recomputeBounds();

        return bounds;
    }

    /**
     * Add a parent to this node.
     *
     * @param p The new parent instance to add to the list
     * @throws AlreadyParentedException There is a valid parent already set
     * @throws InvalidNodeTypeException Not a group node
     */
    protected void addParent(Node p)
        throws AlreadyParentedException, InvalidNodeTypeException
    {
        if(p == null)
            throw new InvalidNodeTypeException(NULL_PARENT_MSG);

        // Check to see that this parent isn't already in the list
        for(int i = 0; i < lastParentList; i++)
            if(parentList[i] == p)
                return;

        resizeList();
        parentList[lastParentList++] = p;
    }

    /**
     * Remove a parent from this shared group. Since setParent() cannot be
     * used to remove a parent from the graph, you'll need to use this method
     * to remove the parent.
     *
     * @param p The new parent instance to remove from the list
     */
    protected void removeParent(Node p)
    {
        // find the location, move everything down one
        for(int i = 0; i < lastParentList; i++)
        {
            if(parentList[i] == p)
            {
                int move_size = lastParentList - i;
                if(move_size != 0)
                    System.arraycopy(parentList,
                                     i,
                                     parentList,
                                     i + 1,
                                     move_size);
                break;
            }
        }
    }

    //---------------------------------------------------------------
    // Methods defined by PickableObjects
    //---------------------------------------------------------------

    /**
     * Set the node as being pickable currently.
     *
     * @param state true to make this node pickable
     */
    public void setPickable(boolean state)
    {
        pickable = state;
    }

    /**
     * Get the current pickable state of this object
     *
     * @return true if this node is currently pickable
     */
    public boolean isPickable()
    {
        return pickable;
    }

    /**
     * Check for all intersections against this geometry to
     * see if there is an intersection with the given set of requests.
     *
     * @param reqs The list of picks to be made, starting at this object
     * @param numRequests The number of valid pick requests to process
     * @throws NotPickableException This object has been marked as non pickable,
     *   but you decided to try to call the method anyway
     * @throws InvalidPickTimingException An attempt was made to pick outside
     *   of the ApplicationUpdateObserver callback method
     */
    public void pickBatch(PickRequest[] reqs, int numRequests)
        throws NotPickableException, InvalidPickTimingException
    {
        if(updateHandler == null || !updateHandler.isPickingPermitted())
            throw new InvalidPickTimingException(PICK_TIMING_MSG);

        if(!pickable)
            throw new NotPickableException(PICKABLE_FALSE_MSG);
    }

    /**
     * Check for all intersections against this geometry to
     * see if there is an intersection with the given set of requests.
     *
     * @param req The details of the pick to be made
     * @throws NotPickableException This object has been marked as non pickable,
     *   but you decided to try to call the method anyway
     * @throws InvalidPickTimingException An attempt was made to pick outside
     *   of the ApplicationUpdateObserver callback method
     */
    public void pickSingle(PickRequest req)
        throws NotPickableException, InvalidPickTimingException
    {
        if(updateHandler == null || !updateHandler.isPickingPermitted())
            throw new InvalidPickTimingException(PICK_TIMING_MSG);

        if(!pickable)
            throw new NotPickableException(PICKABLE_FALSE_MSG);
    }

    //---------------------------------------------------------------
    // Misc Internal methods
    //---------------------------------------------------------------

    /**
     * Check for all intersections against this geometry using a line segment and
     * return the exact distance away of the closest picking point. Default
     * implementation always returns false indicating that nothing was found.
     * Derived classes should override and provide a real implementation.
     *
     * @param start The start point of the segment
     * @param end The end point of the segment
     * @param findAny True if it only has to find a single intersection and can
     *   exit as soon as it finds the first intersection. False if it must find
     *   the closest polygon
     * @param dataOut An array to put the data in for the intersection. Exact
     *   format is described by the flags
     * @param dataOutFlags A set of derived-class specific flags describing what
     *   data should be included in the output array
     * @return True if an intersection was found according to the input request
     * @throws NotPickableException This object has been marked as non pickable,
     *   but you decided to try to call the method anyway
     * @throws InvalidPickTimingException An attempt was made to pick outside
     *   of the ApplicationUpdateObserver callback method
     */
    public boolean pickLineSegment(float[] start,
                                   float[] end,
                                   boolean findAny,
                                   float[] dataOut,
                                   int dataOutFlags)
        throws NotPickableException, InvalidPickTimingException
    {
        if(updateHandler == null || !updateHandler.isPickingPermitted())
            throw new InvalidPickTimingException(PICK_TIMING_MSG);

        if(!pickable)
            throw new NotPickableException(PICKABLE_FALSE_MSG);

        return false;
    }

    /**
     * Check for all intersections against this geometry using a line ray and
     * return the exact distance away of the closest picking point. Default
     * implementation always returns false indicating that nothing was found.
     * Derived classes should override and provide a real implementation.
     *
     * @param origin The start point of the ray
     * @param direction The direction vector of the ray
     * @param findAny True if it only has to find a single intersection and can
     *   exit as soon as it finds the first intersection. False if it must find
     *   the closest polygon
     * @param dataOut An array to put the data in for the intersection. Exact
     *   format is described by the flags
     * @param dataOutFlags A set of derived-class specific flags describing what
     *   data should be included in the output array
     * @return True if an intersection was found according to the input request
     * @throws NotPickableException This object has been marked as non pickable,
     *   but you decided to try to call the method anyway
     * @throws InvalidPickTimingException An attempt was made to pick outside
     *   of the ApplicationUpdateObserver callback method
     */
    public boolean pickLineRay(float[] origin,
                               float[] direction,
                               boolean findAny,
                               float[] dataOut,
                               int dataOutFlags)
        throws NotPickableException, InvalidPickTimingException
    {
        if(updateHandler == null || !updateHandler.isPickingPermitted())
            throw new InvalidPickTimingException(PICK_TIMING_MSG);

        if(!pickable)
            throw new NotPickableException(PICKABLE_FALSE_MSG);

        return false;
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    private final void resizeList()
    {
        if((lastParentList + 1) == parentList.length)
        {
            int old_size = parentList.length;
            int new_size = old_size + LIST_INCREMENT;

            Node[] tmp_nodes = new Node[new_size];

            System.arraycopy(parentList, 0, tmp_nodes, 0, old_size);

            parentList = tmp_nodes;
        }
    }
}