/*****************************************************************************
 *                        Web3d.org Copyright (c) 2001
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// Standard imports
import javax.vecmath.Matrix4f;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * Base representation of a light source.
 * <p>
 *
 * By default a light is not enabled and the colour is set to black. Lights
 * also have a secondary specular value that can be provided. By default it
 * is black (ie has no effect).
 *
 * @author Justin Couch
 * @version $Revision: 1.8 $
 */
public abstract class Light extends Leaf
{
    /** The colour of the light */
    protected float[] diffuseColor;

    /** The colour of the light */
    protected float[] specularColor;

    /** The enabled state */
    protected boolean enabled;

    /**
     * Creates a light with the colour and specular colour set to black.
     */
    protected Light()
    {
        diffuseColor = new float[4];
        diffuseColor[3] = 1;

        specularColor = new float[4];
        specularColor[3] = 1;

        enabled = false;
    }

    /**
     * Create a light with the given base colour.Colour must be in the range
     * [0, 1] otherwise an exception is generated.
     *
     * @param col The new colour value to use
     * @throws IllegalArgumentException The colour value is out of range
     */
    protected Light(float[] col)
        throws IllegalArgumentException
    {
        if(col[0] < 0 || col[0] > 1)
            throw new IllegalArgumentException("Red component out of range");

        if(col[1] < 0 || col[1] > 1)
            throw new IllegalArgumentException("Green component out of range");

        if(col[2] < 0 || col[2] > 1)
            throw new IllegalArgumentException("Blue component out of range");

        diffuseColor = new float[] { col[0], col[1], col[2], 1 };
        enabled = false;
    }

    //---------------------------------------------------------------
    // Misc local methods
    //---------------------------------------------------------------

    /**
     * Set the enabled state of the light. Can use this to turn it on and off
     * in a general fashion.
     *
     * @param state The new state of the light
     */
    public void setEnabled(boolean state)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        enabled = state;
    }

    /**
     * Get the current enabled state of the light.
     *
     * @return The current state
     */
    public boolean isEnabled()
    {
        return enabled;
    }

    /**
     * Set the diffuse colour component to the new value. Colour must be in the
     * range [0, 1] otherwise an exception is generated.
     *
     * @param col The new colour value to use
     * @throws IllegalArgumentException The colour value is out of range
     */
    public void setDiffuseColor(float[] col)
        throws IllegalArgumentException, InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(col[0] < 0 || col[0] > 1)
            throw new IllegalArgumentException("Red component out of range");

        if(col[1] < 0 || col[1] > 1)
            throw new IllegalArgumentException("Green component out of range");

        if(col[2] < 0 || col[2] > 1)
            throw new IllegalArgumentException("Blue component out of range");

        diffuseColor[0] = col[0];
        diffuseColor[1] = col[1];
        diffuseColor[2] = col[2];
    }

    /**
     * Retrieve the current diffuse colour value from the light.
     *
     * @param col An array to copy the colour value into
     */
    public void getDiffuseColor(float[] col)
    {
        col[0] = diffuseColor[0];
        col[1] = diffuseColor[1];
        col[2] = diffuseColor[2];
    }

    /**
     * Set the colour to the new value. Colour must be in the range
     * [0, 1] otherwise an exception is generated.
     *
     * @param col The new colour value to use
     * @throws IllegalArgumentException The colour value is out of range
     */
    public void setSpecularColor(float[] col)
        throws IllegalArgumentException, InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(col[0] < 0 || col[0] > 1)
            throw new IllegalArgumentException("Red component out of range");

        if(col[1] < 0 || col[1] > 1)
            throw new IllegalArgumentException("Green component out of range");

        if(col[2] < 0 || col[2] > 1)
            throw new IllegalArgumentException("Blue component out of range");

        specularColor[0] = col[0];
        specularColor[1] = col[1];
        specularColor[2] = col[2];
    }

    /**
     * Retrieve the current colour value from the light.
     *
     * @param col An array to copy the colour value into
     */
    public void getSpecularColor(float[] col)
    {
        col[0] = specularColor[0];
        col[1] = specularColor[1];
        col[2] = specularColor[2];
    }

    /**
     * Overloaded form of the render() method to render the light details given
     * the specific Light ID used by OpenGL. Since the active light ID for this
     * node may vary over time, a fixed ID cannot be used by OpenGL. The
     * renderer will always call this method rather than the normal render()
     * method. The normal post render will still be called
     *
     * @param gl The GL context to render with
     * @param glu The GL utility context to render with
     * @param lightId the ID of the light to make GL calls with
     */
    public abstract void render(GL gl, GLU glu, int lightId);

    /*
     * Overloaded form of the postRender() method to render the light details given
     * the specific Light ID used by OpenGL. Since the active light ID for this
     * node may vary over time, a fixed ID cannot be used by OpenGL. The
     * renderer will always call this method rather than the normal postRender()
     * method. The normal post render will still be called
     *
     * @param gl The GL context to render with
     * @param glu The GL utility context to render with
     * @param lightId the ID of the light to make GL calls with
     */
    public abstract void postRender(GL gl, GLU glu, int lightId);
}
