/*****************************************************************************
 *                        Web3d.org Copyright (c) 2001
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// Standard imports
import java.util.HashMap;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * An OpenGL LineStripArray.
 *
 * @author Justin Couch
 * @version $Revision: 1.4 $
 */
public class LineStripArray extends VertexGeometry
{
    /** Message when the strip count is < 3 */
    private static final String MIN_COUNT_MSG =
        "A strip with less than 3 vertices is specified at index ";

    /** Message when the total count is less than strip totals */
    private static final String TOTAL_COUNT_MSG =
        "The total requested indexes in the strip is greater than the " +
        "number of valid vertices provided.";

    /** The strip counts of valid triangles for each strip */
    private int[] stripCounts;

    /** The number of valid strips to read from the array */
    private int numStrips;

    /**
     * Constructs a LineStripArray with default values.
     */
    public LineStripArray()
    {
    }

    /**
     * Set the number of valid strips to use. A check is performed to make
     * sure that the number of vertices high enough to support the total
     * of all the strip counts so make sure to call setVertex() with the
     * required array length before calling this method. Each strip must be
     * a minumum length of three.
     *
     * @param counts The array of counts
     * @param num The number of valid items to read from the array
     * @throws IllegalArgumentException Invalid total strip count or
     *   individual strip count < 3
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setStripCount(int[] counts, int num)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        int total = 0;

        for(int i = 0; i < num; i++)
        {
            if(counts[i] < 3)
                throw new IllegalArgumentException(MIN_COUNT_MSG + i);

            total += counts[i];
        }

        if(stripCounts == null || stripCounts.length < num)
            stripCounts = new int[num];

        if(total > numCoords)
            throw new IllegalArgumentException(TOTAL_COUNT_MSG);

        numStrips = num;

        if(num > 0)
            System.arraycopy(counts, 0, stripCounts, 0, num);

        stateChanged = true;
    }

    /**
     * Get the number of valid strips that are defined for this geometry.
     *
     * @return a positive number
     */
    public int getValidStripCount()
    {
        return numStrips;
    }

    /**
     * Get the sizes of the valid strips. The passed array must be big enough
     * to contain all the strips.
     *
     * @param values An array to copy the strip values into
     */
    public void getStripCount(int[] values)
    {
        System.arraycopy(stripCounts, 0, values, 0, numStrips);
    }

    /**
     * Issue ogl commands needed for this renderable object.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void renderState(GL gl, GLU glu)
    {
        // If we have changed state, then clear the old display lists
        if(stateChanged)
        {
            synchronized(displayListMap)
            {
                displayListMap.clear();
                stateChanged = false;
            }
        }

        // No coordinates, do nothing.
        if((numStrips == 0) || ((vertexFormat & COORDINATE_MASK) == 0))
            return;

        Integer listName = (Integer)displayListMap.get(gl);

        if(listName == null)
        {
            listName = new Integer(gl.glGenLists(1));

            gl.glNewList(listName.intValue(), GL.GL_COMPILE_AND_EXECUTE);

            //  State enable first
            int vtx_size = vertexFormat & 0x07;
            gl.glEnableClientState(GL.GL_VERTEX_ARRAY);
            gl.glVertexPointer(vtx_size, GL.GL_FLOAT, 0, vertexBuffer);

            if((vertexFormat & NORMALS) != 0)
            {
                gl.glEnableClientState(GL.GL_NORMAL_ARRAY);
                gl.glNormalPointer(GL.GL_FLOAT, 0, normalBuffer);
            }

            if((vertexFormat & TEXTURE_MASK) != 0)
            {
                // single texturing or multi-texturing
                if(numTextureSets == 1)
                {
                    gl.glEnableClientState(GL.GL_TEXTURE_COORD_ARRAY);
                    gl.glTexCoordPointer(textureTypes[0],
                                         GL.GL_FLOAT,
                                         0,
                                         textureBuffer[0]);
                }
                else
                {
                    for(int i = 0; i < numTextureSets; i++)
                    {
                        int set_id = textureSets[i];
                        gl.glClientActiveTexture(GL.GL_TEXTURE0 + i);
                        gl.glEnableClientState(GL.GL_TEXTURE_COORD_ARRAY);
                        gl.glTexCoordPointer(textureTypes[set_id],
                                             GL.GL_FLOAT,
                                             0,
                                             textureBuffer[set_id]);
                    }
                }
            }

            if((vertexFormat & COLOR_MASK) != 0)
            {
                gl.glEnableClientState(GL.GL_COLOR_ARRAY);
                int size = ((vertexFormat & COLOR_3) != 0) ? 3 : 4;
                gl.glColorPointer(size, GL.GL_FLOAT, 0, colorBuffer);
            }

            // Loop through the provided arrays using the sub-range
            // draw capabilities of glDrawArrays to create multiple
            // strips being drawn.
            int strip_offset = 0;
            for(int i = 0; i < numStrips; i++)
            {
                gl.glDrawArrays(GL.GL_LINE_STRIP,
                                strip_offset,
                                stripCounts[i]);

                strip_offset += stripCounts[i];
            }
            gl.glDisableClientState(GL.GL_VERTEX_ARRAY);

            if((vertexFormat & NORMALS) != 0)
                gl.glDisableClientState(GL.GL_NORMAL_ARRAY);

            if((vertexFormat & TEXTURE_MASK) != 0)
                gl.glDisableClientState(GL.GL_TEXTURE_COORD_ARRAY);

            if((vertexFormat & COLOR_MASK) != 0)
                gl.glDisableClientState(GL.GL_COLOR_ARRAY);

            gl.glEndList();
            displayListMap.put(gl, listName);
        }
        else
        {
            gl.glCallList(listName.intValue());
        }
    }

    /**
     * Restore all openGL state.
     *
      * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void restoreState(GL gl, GLU glu)
    {
    }
}