/*****************************************************************************
 *                        Web3d.org Copyright (c) 2001
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// Standard imports
import java.util.HashMap;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * Describes the material properties of an object.
 *
 * @author Alan Hudson
 * @version $Revision: 1.10 $
 */
public class Material extends NodeComponent
{
    /** Default diffuse colour */
    private static final float[] DEFAULT_DIFFUSE = {0.8f, 0.8f, 0.8f};

    /** Default ambient colour */
    private static final float[] DEFAULT_AMBIENT = {0.2f, 0.2f, 0.2f};

    /** Default emissive colour */
    private static final float[] DEFAULT_EMISSIVE = {0, 0, 0};

    /** Default specular colour */
    private static final float[] DEFAULT_SPECULAR = {0, 0, 0};

    /** Default shininess colour */
    private static final float DEFAULT_SHININESS = 0.2f;

    /** Default transparency value */
    private static final float DEFAULT_TRANSPARENCY = 1;

    /** Should lighting be used for this material? */
    private boolean useLighting;

    /** Diffuse colour components */
    private float[] diffuseColor;

    /** Ambient colour components */
    private float[] ambientColor;

    /** Emissive colour components */
    private float[] emissiveColor;

    /** Specular colour components */
    private float[] specularColor;

    /** shinine colour factor */
    private float shininess;

    /** A mapping between glContext and displayListID(Integer) */
    private HashMap displayListMap;

    /** Flag to say the state has changed in the material */
    private boolean stateChanged;

    /**
     * Constructs a material with default values.
     * <pre>
     * ambient color = (0.2, 0.2, 0.2)
     * emisive color = (0.0, 0.0, 0.0)
     * diffuse color = (0.8, 0.8, 0.8)
     * specular color = (0, 0, 0)
     * shininess = 0.2
     * transparency = 1
     * </pre>
     */
    public Material()
    {
        this(DEFAULT_AMBIENT,
             DEFAULT_EMISSIVE,
             DEFAULT_DIFFUSE,
             DEFAULT_SPECULAR,
             DEFAULT_SHININESS,
             DEFAULT_TRANSPARENCY);
    }

    /**
     * Create a new material with all the colours specified.
     *
     * @param ambientColor
     * @param emissiveColor
     * @param diffuseColor
     * @param specularColor
     * @param shininess The shininess factor value
     * @param transparency A transparency value between 0 and 1
     */
    public Material(float[] ambientColor,
                    float[] emissiveColor,
                    float[] diffuseColor,
                    float[] specularColor,
                    float shininess,
                    float transparency)
    {
        this.diffuseColor = new float[4];
        this.diffuseColor[0] = diffuseColor[0];
        this.diffuseColor[1] = diffuseColor[1];
        this.diffuseColor[2] = diffuseColor[2];
        this.diffuseColor[3] = transparency;

        this.ambientColor = new float[3];
        this.ambientColor[0] = ambientColor[0];
        this.ambientColor[1] = ambientColor[1];
        this.ambientColor[2] = ambientColor[2];

        this.specularColor = new float[3];
        this.specularColor[0] = specularColor[0];
        this.specularColor[1] = specularColor[1];
        this.specularColor[2] = specularColor[2];

        this.emissiveColor = new float[3];
        this.emissiveColor[0] = emissiveColor[0];
        this.emissiveColor[1] = emissiveColor[1];
        this.emissiveColor[2] = emissiveColor[2];

        // Convert [0,1] to [0,128] for openGL
        this.shininess = 128 * shininess;
        useLighting = true;

        displayListMap = new HashMap(1);
        stateChanged = true;

    }

    /**
     * Set the shininess factor.
     *
     * @param s The shininess factor value
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setShininess(float s)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        // Convert [0,1] to [0,128] for openGL
        shininess = 128 * s;
        stateChanged = true;
    }

    /**
     * Set the diffuse color to the new value.
     *
     * @param col The new colour to use
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setDiffuseColor(float[] col)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        diffuseColor[0] = col[0];
        diffuseColor[1] = col[1];
        diffuseColor[2] = col[2];
        stateChanged = true;
    }

    /**
     * Set the specular color to the new value.
     *
     * @param col The new colour to use
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setSpecularColor(float[] col)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        specularColor[0] = col[0];
        specularColor[1] = col[1];
        specularColor[2] = col[2];
        stateChanged = true;
    }

    /**
     * Set the emissive color to the new value.
     *
     * @param col The new colour to use
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setEmissiveColor(float[] col)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        emissiveColor[0] = col[0];
        emissiveColor[1] = col[1];
        emissiveColor[2] = col[2];
        stateChanged = true;
    }

    /**
     * Set the ambient color to the new value.
     *
     * @param col The new colour to use
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setAmbientColor(float[] col)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        ambientColor[0] = col[0];
        ambientColor[1] = col[1];
        ambientColor[2] = col[2];
        stateChanged = true;
    }

    /**
     * Set the amount of transparency to be used for this material.
     *
     * @param transparency A value between 0 and 1
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setTransparency(float transparency)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        diffuseColor[3] = transparency;
        stateChanged = true;
    }

    /**
     * Get the value for the existing transparency values
     *
     * @return The current transparency setting
     */
    public float getTransparency()
    {
        return diffuseColor[3];
    }

    /**
     * Set the flag that enables or disables lighting on any geometry using
     * this material.
     *
     * @param state true if lighting should be used
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setLightingEnabled(boolean state)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(useLighting != state)
        {
            useLighting = state;
            stateChanged = true;
        }
    }

    /**
     * Get the current lighting enabled flag.
     */
    public boolean isLightingEnabled()
    {
        return useLighting;
    }

    /**
     * Issue ogl commands needed for this component
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void renderState(GL gl, GLU glu)
    {
        // If we have changed state, then clear the old display lists
        if(stateChanged)
        {
            synchronized(displayListMap)
            {
                displayListMap.clear();
                stateChanged = false;
            }
        }

        Integer listName = (Integer)displayListMap.get(gl);

        if(listName == null)
        {
            listName = new Integer(gl.glGenLists(1));

            gl.glNewList(listName.intValue(), GL.GL_COMPILE_AND_EXECUTE);

            if(useLighting)
                gl.glEnable(GL.GL_LIGHTING);

            gl.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_DIFFUSE, diffuseColor);
            gl.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_AMBIENT, ambientColor);
            gl.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_SPECULAR, specularColor);
            gl.glMaterialfv(GL.GL_FRONT_AND_BACK, GL.GL_EMISSION, emissiveColor);
            gl.glMaterialf(GL.GL_FRONT_AND_BACK, GL.GL_SHININESS, shininess);

            gl.glEndList();
            displayListMap.put(gl, listName);
        }
        else
        {
            gl.glCallList(listName.intValue());
        }
    }

    /**
     * Restore all openGL state to the given drawable
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void restoreState(GL gl, GLU glu)
    {
    }
}