/*****************************************************************************
 *                        Web3d.org Copyright (c) 2003
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// Standard imports
import java.util.HashMap;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * Describes attributes used when rendering a polygon.
 * <p>
 *
 *
 * @author Justin Couch
 * @version $Revision: 1.8 $
 */
public class PolygonAttributes extends NodeComponent
{
    /** Message when the drawing mode is not a valid value */
    private static final String INVALID_DRAW_MSG =
        "An invalid draw mode was provided. It must be one of " +
        "DRAW_POINT, DRAW_LINE or DRAW_FILLED";


    /** Message when the face cull mode is not a valid value */
    private static final String INVALID_CULL_MSG =
        "An invalid face cull mode was provided. It must be one of " +
        "CULL_NONE, CULL_FRONT, CULL_BACK or CULL_BOTH";

    /** Message when the face cull mode is not a valid value */
    private static final String STIPPLE_SIZE_MSG =
        "The stipple pattern array must be at least 1024 items in size";

    /** The cull mode says to draw both front and back faces */
    public static final int CULL_NONE = 0;

    /** The cull mode says to draw front faces only */
    public static final int CULL_FRONT = GL.GL_FRONT;

    /** The cull mode says to draw back faces only */
    public static final int CULL_BACK = GL.GL_BACK;

    /** The cull mode says to not draw any faces */
    public static final int CULL_BOTH = GL.GL_FRONT_AND_BACK;

    /** Draw the face as the points of the vertices only */
    public static final int DRAW_POINT = GL.GL_POINT;

    /** Draw the face as outline lines only */
    public static final int DRAW_LINE = GL.GL_LINE;

    /** Draw the face filled as a solid object */
    public static final int DRAW_FILLED = GL.GL_FILL;

    /** The rendering draw mode for the front face */
    private int frontDrawMode;

    /** The rendering draw mode for the back face */
    private int backDrawMode;

    /** Winding direction for the polygon */
    private boolean isCCW;

    /** Cull operations to be performed */
    private int cullFace;

    /** The polygon offset factor amount */
    private float polyOffsetFactor;

    /** The polygon offset unit amount */
    private float polyOffsetUnit;

    /** Has an attribute changed */
    private boolean stateChanged;

    /** Map of display contexts to maps */
    private HashMap displayListMap;

    /** Stipple pattern for the polygon, if desired */
    private byte[] stipple;

    /**
     * Constructs a attribute set with default values.
     */
    public PolygonAttributes()
    {
        stateChanged = true;
        isCCW = true;
        cullFace = CULL_NONE;
        polyOffsetFactor = 0;
        polyOffsetUnit = 1;

        frontDrawMode = DRAW_FILLED;
        backDrawMode = DRAW_FILLED;

        displayListMap = new HashMap(1);
    }

    /**
     * Issue ogl commands needed for this component
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void renderState(GL gl, GLU glu)
    {
        // If we have changed state, then clear the old display lists
        if(stateChanged)
        {
            synchronized(displayListMap)
            {
                displayListMap.clear();
                stateChanged = false;
            }
        }

        Integer listName = (Integer)displayListMap.get(gl);

        if(listName == null)
        {
            listName = new Integer(gl.glGenLists(1));

            gl.glNewList(listName.intValue(), GL.GL_COMPILE_AND_EXECUTE);

            gl.glPushAttrib(GL.GL_POLYGON_BIT);

            if(!isCCW)
                gl.glFrontFace(GL.GL_CW);

            if(cullFace != CULL_NONE)
                gl.glCullFace(cullFace);

            // Using polygon offsets requires that the polygon mode be
            // explicitly set, but we don't want to set it if we don't
            // have to.
            if(polyOffsetFactor != 0)
            {
                gl.glEnable(GL.GL_POLYGON_OFFSET_FILL);
                gl.glPolygonOffset(polyOffsetFactor, polyOffsetUnit);

                gl.glPolygonMode(GL.GL_FRONT, frontDrawMode);
                gl.glPolygonMode(GL.GL_BACK, backDrawMode);
            }
            else
            {
                if(frontDrawMode != DRAW_FILLED)
                    gl.glPolygonMode(GL.GL_FRONT, frontDrawMode);

                if(backDrawMode != DRAW_FILLED)
                    gl.glPolygonMode(GL.GL_BACK, backDrawMode);
            }

            if(stipple != null)
            {
                gl.glEnable(GL.GL_POLYGON_STIPPLE);
                gl.glPixelStorei(GL.GL_UNSIGNED_BYTE, 1);
                gl.glPolygonStipple(stipple);
            }

            gl.glEndList();
            displayListMap.put(gl, listName);
        }
        else
        {
            gl.glCallList(listName.intValue());
        }
    }

    /**
     * Restore all openGL state to the given drawable.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void restoreState(GL gl, GLU glu)
    {
        if(polyOffsetFactor != 0)
            gl.glDisable(GL.GL_POLYGON_OFFSET_FILL);

        if(stipple != null)
            gl.glDisable(GL.GL_POLYGON_STIPPLE);

        gl.glPopAttrib();
    }

    /**
     * Set the CCW flag. By default CCW is true.
     *
     * @param state True to use CCW triangles, false for CW
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setCCW(boolean state)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(isCCW != state)
        {
            isCCW = state;
            stateChanged = true;
        }
    }

    /**
     * Check the state of the CCW flag setting for this geometry.
     *
     * @return true if the vertices are rendered counter clockwise
     */
    public boolean isCCW()
    {
        return isCCW;
    }

    /**
     * Set which face is to be culled. It should be one of the base
     * constants defined by this class.
     *
     * @param face The face that should be culled
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setCulledFace(int face)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        switch(face)
        {
            case CULL_NONE:
            case CULL_FRONT:
            case CULL_BACK:
            case CULL_BOTH:
                if(face != cullFace)
                {
                    cullFace = face;
                    stateChanged = true;
                }
                break;

            default:
                throw new IllegalArgumentException(INVALID_CULL_MSG);
        }
    }

    /**
     * Get the current face culled flag.
     *
     * @return one of the four CULL_X constants
     */
    public int getCulledFace()
    {
        return cullFace;
    }

    /**
     * Set the draw mode for either the front or back face.
     *
     * @param front true if this is the front-face setting
     * @param mode The mode to use for this face
     * @throws IllegalArgumentException Invalid mode supplied
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setDrawMode(boolean front, int mode)
        throws IllegalArgumentException, InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        switch(mode)
        {
            case DRAW_POINT:
            case DRAW_LINE:
            case DRAW_FILLED:
                if(front)
                {
                    if(mode != frontDrawMode)
                    {
                        frontDrawMode = mode;
                        stateChanged = true;
                    }
                }
                else
                {
                    if(mode != backDrawMode)
                    {
                        backDrawMode = mode;
                        stateChanged = true;
                    }
                }
                break;

            default:
                throw new IllegalArgumentException(INVALID_DRAW_MSG);
        }
    }

    /**
     * Get the draw mode for the specified face.
     *
     * @param front true if requesting the front-face setting
     */
    public int getDrawMode(boolean front)
    {
        return front ? frontDrawMode : backDrawMode;
    }

    /**
     * Set the polygon offset details. Factor and unit applied as per the
     * OpenGL specification.
     *
     * @param factor The offset factor mulitplier
     * @param units The offset unit multiplier
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setPolygonOffset(float factor, float units)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(polyOffsetFactor != factor)
        {
            polyOffsetFactor = factor;
            stateChanged = true;
        }

        if(polyOffsetUnit != units)
        {
            polyOffsetUnit = units;
            stateChanged = true;
        }
    }

    /**
     * Get the current polygon offset details. The array is used to copy the
     * current values into and therefore must be at least 2 items in length.
     * Index 0 is the factor, index 1 is the units
     *
     * @param values The array to copy the current values into
     */
    public void getPolygonOffset(float[] values)
    {
        values[0] = polyOffsetFactor;
        values[1] = polyOffsetUnit;
    }

    /**
     * Set the stipple mask to be used on the polygon. If the stipple is null
     * then it will clear the current stipple setup. If not null, it must be
     * at least an array length 1024 (32x32). The value is used by reference
     *
     *
     * @param pattern The bytes of the pattern or null
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setStipplePattern(byte[] pattern)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(pattern != null)
        {
            if(pattern.length < 1024)
                throw new IllegalArgumentException(STIPPLE_SIZE_MSG);

            stipple = pattern;
        }
        else
        {
            stipple = null;
        }
    }
}