/*****************************************************************************
 *                        Web3d.org Copyright (c) 2001
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// External imports
// None

// Local imports
// None

/**
 * A marker interface that represents a class capable of managing
 * the complete rendering pipeline.
 * <p>
 *
 * A manager is used to handle a system-specific rendering technique. The goal
 * is to manage the contained pipeline(s) in a way that is most efficient to
 * the hardware provided. Thus, it is expected there will be many different
 * types of managers to suit the many hardware configurations available.
 * <p>
 *
 * Example implementations of the pipeline manager would be one that handles
 * all the pipelines with simultaneous threads, each pinned to a particular
 * CPU/Graphics pipe that the machine has. Another implementation may hold all
 * the pipelines for sequential evaluation piping the output from one into the
 * input for another (eg for handling dynamic cubic environment maps).
 * <p>
 *
 * Basic common methods are provided for all implementations to use. It is
 * expected that implementations will add additional technique-specific
 * extension methods to the basic features.
 *
 * @author Justin Couch
 * @version $Revision: 1.2 $
 */
public class Scene
{
    /** User-defined data storage */
    private Object userData;

    /** Counter to know whether this scene is live or not */
    private int liveCount;

    /** Simple liveness state to accompany the counter */
    private boolean alive;

    /** The Scene Graph renderableObjects */
    private Group renderableObjects;

    /** The current viewpoint instance */
    private Viewpoint currentViewpoint;

    /** The current fog instance */
    private Fog currentFog;

    /** The current background instance */
    private Background currentBackground;

    /** The current clear colour */
    private float[] clearColor;

    /** Current view environment */
    private ViewEnvironment viewEnvironment;

    /** Current effects processor */
    private RenderEffectsProcessor processor;

    /**
     * Create a default instance of this scene with no content provided.
     */
    public Scene()
    {
        liveCount = 0;
        alive = false;
    }

    /**
     * Notification that this object is live now. Overridden to make sure that
     * the live state of the nodes represents the same state as the parent
     * scene graph.
     *
     * @param state true if this should be marked as live now
     */
    protected void setLive(boolean state)
    {
        if(state)
            liveCount++;
        else if(liveCount > 0)
            liveCount--;

        if((liveCount == 0) || !alive)
        {
            alive = state;

            if(renderableObjects != null)
                renderableObjects.setLive(state);
        }
    }

    /**
     * Register the scene processor to be used for this scene for pre and
     * post rendering effects. Setting a null value will clear the current
     * instance.
     *
     * @param prc The instance to use or null
     */
    public void setRenderEffectsProcessor(RenderEffectsProcessor prc)
    {
        processor = prc;
    }

    /**
     * Get the currently set scene processor instance. If none is set,
     * returns null.
     *
     * @return The current processor or null
     */
    public RenderEffectsProcessor getRenderEffectsProcessor()
    {
        return processor;
    }

    /**
     * Assign some arbitrary piece of user data to this scene. This data
     * is also sent back to the user through scene processor callbacks when
     * they are made. Setting a null value will clear the current instance.
     *
     * @param data some arbitrary data the user wants to provide
     */
    public void setUserData(Object data)
    {
        userData = data;
    }

    /**
     * Get the currently set instance of user data. If none provided, return
     * null.
     *
     * @return Some arbitrary piece of data or null
     */
    public Object getUserData()
    {
        return userData;
    }

    /**
     * Set the collection of geometry that should be rendered to this
     * texture.
     *
     * A null value will clear the current geometry and result in only
     * rendering the background, if set. if not set, then whatever the default
     * colour is, is used (typically black).
     *
     * @param geom The new geometry to use or null
     */
    public void setRenderedGeometry(Group geom)
    {
        if(renderableObjects != null)
            renderableObjects.setLive(false);

        renderableObjects = geom;

        if(renderableObjects != null)
            renderableObjects.setLive(alive);
    }

    /**
     * Get the root of the currently rendered scene. If none is set, this will
     * return null.
     *
     * @return The current scene root or null.
     */
    public Group getRenderedGeometry()
    {
        return renderableObjects;
    }

    /**
     * Set the viewpoint path that should be applied to the current surface.
     * The output drawn will be a combination of this information and that
     * of the view environment.
     *
     * @param vp The instance of the active viewpoint to use
     * @throws IllegalArgumentException The path contains a SharedGroup or
     *    the node is not live
     */
    public void setActiveView(Viewpoint vp)
        throws IllegalArgumentException
    {
        currentViewpoint = vp;
    }

    /**
     * Get the currently set active view. If none is set, return null.
     *
     * @return The current view instance or null
     */
    public Viewpoint getActiveView()
    {
        return currentViewpoint;
    }

    /**
     * Set the fog that should be applied to the current surface.
     * The output drawn will be a combination of this information and that
     * of the view environment.
     *
     * @param fog The instance of the active fog node
     * @throws IllegalArgumentException The path contains a SharedGroup or
     *    the node is not live
     */
    public void setActiveFog(Fog fog)
        throws IllegalArgumentException
    {
        currentFog = fog;
    }

    /**
     * Get the currently set active view. If none is set, return null.
     *
     * @return The current view instance or null
     */
    public Fog getActiveFog()
    {
        return currentFog;
    }

    /**
     * Set the background path that should be applied to the current surface.
     * The output drawn will be a combination of this information and that
     * of the view environment.
     *
     * @param bg The instance of the active background
     * @throws IllegalArgumentException The path contains a SharedGroup or
     *    the node is not live
     */
    public void setActiveBackground(Background bg)
        throws IllegalArgumentException
    {
        currentBackground = bg;
    }

    /**
     * Get the currently set active view. If none is set, return null.
     *
     * @return The current view instance or null
     */
    public Background setActiveBackground()
    {
        return currentBackground;
    }

    /**
     * Set the view environment that is used to render this surface.
     *
     * @param env The environment instance to use for the render setup
     */
    public void setViewEnvironment(ViewEnvironment env)
    {
        viewEnvironment = env;
    }

    /**
     * Get the currently set active view. If none is set, return null.
     *
     * @return The current view instance or null
     */
    public ViewEnvironment getViewEnvironment()
    {
        return viewEnvironment;
    }
}