/*****************************************************************************
 *                        Web3d.org Copyright (c) 2003
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// Standard imports
import java.util.HashMap;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * Describes a texture's automatic texture coordinate generation.
 * <p>
 *
 * Texture modes here directly correspond to the OpenGL constants
 * of the same type. Either are acceptable as parameters.
 *
 * @author Justin Couch
 * @version $Revision: 1.6 $
 */
public class TexCoordGeneration extends NodeComponent
{
    /** Message when an invalid coordinate is given */
    private static final String INVALID_COORD_MSG =
        "The coordinate specified is not one of S,T,R or Q";

    /** Generate coordinates for a texture's S coordinate */
    public static final int TEXTURE_S = GL.GL_S;

    /** Generate coordinates for a texture's T coordinate  */
    public static final int TEXTURE_T = GL.GL_T;

    /** Generate coordinates for a texture's R coordinate */
    public static final int TEXTURE_R = GL.GL_R;

    /** Generate coordinates for a texture's Q coordinate  */
    public static final int TEXTURE_Q = GL.GL_Q;

    /** Coordinate reference plane is user defined */
    public static final int MODE_GENERIC = GL.GL_TEXTURE_GEN_MODE;

    /** Generate reference plane is relative to the object */
    public static final int MODE_OBJECT_PLANE = GL.GL_OBJECT_PLANE;

    /** Generate reference plane is relative to the eye position */
    public static final int MODE_EYE_PLANE = GL.GL_EYE_PLANE;

    /** Generate coordinates in a simple linear fashion on the object */
    public static final int MAP_OBJECT_LINEAR = GL.GL_OBJECT_LINEAR;

    /** Generate coordinates in a simple linear fashion on the eye position */
    public static final int MAP_EYE_LINEAR = GL.GL_EYE_LINEAR;

    /** Generate coordinates in a spherical shape for env mapping */
    public static final int MAP_SPHERICAL = GL.GL_SPHERE_MAP;

    /**
     * Generate coordinates using the normals at the vertex. Used principly by
     * cubic environment mapping.
     */
    public static final int MAP_NORMALS = GL.GL_NORMAL_MAP;

    /**
     * Generate coordinates using the normals at the vertex. Used principly by
     * cubic environment mapping.
     */
    public static final int MAP_REFLECTIONS = GL.GL_REFLECTION_MAP;

    /** Modes that are needed to be passed to the system for tex gen */
    private static final int[] ENABLE_MODE =
    {
        GL.GL_TEXTURE_GEN_S,
        GL.GL_TEXTURE_GEN_T,
        GL.GL_TEXTURE_GEN_R,
        GL.GL_TEXTURE_GEN_Q,
    };

    /** Parameters used when TEXTURE_GEN or OBJECT_LINEAR used */
    private float[][] parameters;

    /** The texture coordiante effected */
    private int[] coordinate;

    /** The parameter mode use */
    private int[] modes;

    /** The mapping type to use */
    private int[] mapping;

    /** A mapping between glContext and displayListID(Integer) */
    private HashMap displayListMap;

    /** Flag to say the state has changed in the material */
    private boolean stateChanged;

    /**
     * Constructs a TexCoordGeneration with default values, which is
     * to say, do nothing.
     */
    public TexCoordGeneration()
    {
        parameters = new float[4][];
        coordinate = new int[4];
        modes = new int[4];
        mapping = new int[4];

        displayListMap = new HashMap();
        stateChanged = false;
    }

    /**
     * Create automatic coordinate generation for one axis with
     * the given set of abilities.
     *
     * @throws IllegalArgumentException Invalid axis, mode or parameter
     */
    public TexCoordGeneration(int axis, int mode, int parameter, float[] value)
    {
        this();

        setParameter(axis, mode, parameter, value);
    }

    /**
     * Setup one of the axis parameters.
     *
     */
    public void setParameter(int axis, int mode, int parameter, float[] value)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        switch(axis)
        {
            case TEXTURE_S:
                coordinate[0] = TEXTURE_S;
                modes[0] = mode;
                mapping[0] = parameter;
                if(value != null)
                {
                    if(parameters[0] == null)
                        parameters[0] = new float[4];

                    parameters[0][0] = value[0];
                    parameters[0][1] = value[1];
                    parameters[0][2] = value[2];
                    parameters[0][3] = value[3];
                }

                stateChanged = true;
                break;

            case TEXTURE_T:
                coordinate[1] = TEXTURE_T;

                modes[1] = mode;
                mapping[1] = parameter;
                if(value != null)
                {
                    if(parameters[1] == null)
                        parameters[1] = new float[4];

                    parameters[1][0] = value[0];
                    parameters[1][1] = value[1];
                    parameters[1][2] = value[2];
                    parameters[1][3] = value[3];
                }
                stateChanged = true;
                break;

            case TEXTURE_R:
                coordinate[2] = TEXTURE_R;
                modes[2] = mode;
                mapping[2] = parameter;
                if(value != null)
                {
                    if(parameters[2] == null)
                        parameters[2] = new float[4];

                    parameters[2][0] = value[0];
                    parameters[2][1] = value[1];
                    parameters[2][2] = value[2];
                    parameters[2][3] = value[3];
                }
                stateChanged = true;
                break;

            case TEXTURE_Q:
                coordinate[3] = TEXTURE_Q;
                modes[3] = mode;
                mapping[3] = parameter;
                if(value != null)
                {
                    if(parameters[3] == null)
                        parameters[3] = new float[4];

                    parameters[3][0] = value[0];
                    parameters[3][1] = value[1];
                    parameters[3][2] = value[2];
                    parameters[3][3] = value[3];
                }
                stateChanged = true;
                break;

            default:
                throw new IllegalArgumentException(INVALID_COORD_MSG);
        }
    }

    /**
     * Issue ogl commands needed for this component
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void renderState(GL gl, GLU glu)
    {
        // If we have changed state, then clear the old display lists
        if(stateChanged)
        {
            synchronized(displayListMap)
            {
                displayListMap.clear();
                stateChanged = false;
            }
        }

        Integer listName = (Integer)displayListMap.get(gl);

        if(listName == null)
        {
            listName = new Integer(gl.glGenLists(1));

            gl.glNewList(listName.intValue(), GL.GL_COMPILE_AND_EXECUTE);

            //  State enable first
            for(int i = 0; i < 4; i++)
            {
                if(coordinate[i] == 0)
                    continue;

                if(modes[i] == MODE_GENERIC)
                {
                    gl.glTexGeni(coordinate[i], modes[i], mapping[i]);

                    if(mapping[i] == MAP_OBJECT_LINEAR)
                        gl.glTexGenfv(coordinate[i],
                                      MODE_OBJECT_PLANE,
                                      parameters[i]);
                }
                else
                {
                    gl.glTexGenfv(coordinate[i], modes[i], parameters[i]);
                }

                gl.glEnable(ENABLE_MODE[i]);
            }

            gl.glEndList();
            displayListMap.put(gl, listName);
        }
        else
        {
            gl.glCallList(listName.intValue());
        }
    }

    /**
     * Restore all openGL state to the given drawable
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void restoreState(GL gl, GLU glu)
    {
    }
}