/*****************************************************************************
 *                        Web3d.org Copyright (c) 2001
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// Standard imports
import java.util.HashMap;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLDrawable;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * Describes the texture properties of an object.
 *
 * @author Alan Hudson
 * @version $Revision: 1.13 $
 */
public abstract class Texture extends NodeComponent
    implements DeletableSceneGraphObject
{
    /** MipMapMode constants - No Mip Map*/
    public static final int MODE_BASE_LEVEL = 0;

    /** MipMapMode constants - Use Mip Maps */
    public static final int MODE_MIPMAP = 1;

    // Boundary Modes
    /** Boundary mode to repeat textures */
    public static final int BM_WRAP = GL.GL_REPEAT;

    /** Boundary mode to clamp textures */
    public static final int BM_CLAMP = GL.GL_CLAMP;

    /** Boundary mode to clamp the texture edge value without border */
    public static final int BM_CLAMP_TO_EDGE = GL.GL_CLAMP_TO_EDGE;

    /** Boundary mode to clamp the texture border colour */
    public static final int BM_CLAMP_TO_BOUNDARY = GL.GL_CLAMP_TO_BORDER;
    //public static final int BM_MIRRORED_REPEAT = GL.GL_MIRRORED_REPEAT;

    // Maxification Filter Techniques

    /** Set the maginification filter to the fastest option */
    public static final int MAGFILTER_FASTEST = 0;

    /** Set the maginification filter to the highest quality option */
    public static final int MAGFILTER_NICEST = 1;

    /** Set the maginification filter to base point-base filtering */
    public static final int MAGFILTER_BASE_LEVEL_POINT = 2;

    /** Set the maginification filter to linear filtering */
    public static final int MAGFILTER_BASE_LEVEL_LINEAR = 3;

    /** Set the maginfication filter to use the detail texture option */
    public static final int MAGFILTER_LINEAR_DETAIL = 4;

    /** Set the maginfication filter to use the detail texture's RGB values */
    public static final int MAGFILTER_LINEAR_DETAIL_RGB = 5;

    /** Set the maginfication filter to use the detail texture's alpha values */
    public static final int MAGFILTER_LINEAR_DETAIL_ALPHA = 6;

    // Minification Filter Techniques
    /** Set the minification filter to the fastest option */
    public static final int MINFILTER_FASTEST = 0;

    /** Set the mininification filter to the highest quality  option */
    public static final int MINFILTER_NICEST = 1;

    /** Set the mininification filter to base point-base filtering */
    public static final int MINFILTER_BASE_LEVEL_POINT = 2;

    /** Set the mininification filter to linear filtering */
    public static final int MINFILTER_BASE_LEVEL_LINEAR = 3;

    /** Set the mininification filter to base point-base filtering */
    public static final int MINFILTER_MULTI_LEVEL_POINT = 4;

    /** Set the mininification filter to linear filtering */
    public static final int MINFILTER_MULTI_LEVEL_LINEAR = 5;

    // Anistropic Modes

    /** Disable anisotropic filtering */
    public static final int ANISOTROPIC_MODE_NONE = 0;

    /** Enable anisotropic filtering */
    public static final int ANISOTROPIC_MODE_SINGLE = 1;

    // Format values
    // More need to be added here later.

    /** Interpret the texture format as alpha only */
    public static final int FORMAT_ALPHA = GL.GL_ALPHA;

    /** Interpret the texture format as intensity only */
    public static final int FORMAT_INTENSITY = GL.GL_LUMINANCE;

    /** Interpret the texture format as intensity-alpha */
    public static final int FORMAT_INTENSITY_ALPHA = GL.GL_LUMINANCE_ALPHA;

    /** Interpret the texture format as RGB */
    public static final int FORMAT_RGB = GL.GL_RGB;

    /** Interpret the texture format as RGBA */
    public static final int FORMAT_RGBA = GL.GL_RGBA;

    /** The images defining this texture */
    protected TextureComponent[] images;

    /** The number of valid items in the image array */
    protected int numImages;

    /** The Anisotropic Filtering Mode */
    protected int anisotropicMode;

    /** The Anisotropic Filtering Degree */
    protected float anisotropicDegree;

    /** The magnification filter */
    protected int magFilter;

    /** The minification filter */
    protected int minFilter;

    /** The boundary mode S value */
    protected int boundaryModeS;

    /** The mipMapMode */
    protected int mipMapMode;

    /** The width of the main texture. */
    protected int width;

    /** The pixel format of the main texture image. */
    protected int format;

    /**
     * Flag to say that the display lists must be cleared and regenerated
     * because some state changed
     */
    protected boolean stateChanged;

    /** A mapping between glContext and displayListID(Integer) */
    protected HashMap displayListMap;

    /** The GL type of the texture. */
    protected final int textureType;

    /** The OpenGL ID of this texture */
    protected int textureId;

    /**
     * Constructs a texture with default values. The mipmap mode is set to
     * MODE_BASE_LEVEL.
     */
    public Texture(int type)
    {
        width = -1;
        textureId = -1;
        textureType = type;
        mipMapMode = MODE_BASE_LEVEL;

        displayListMap = new HashMap(1);
    }

    /**
     * Set the images for this texture
     *
     * @param mipMapMode Flag stating the type of texture mode to use
     * @param format Image format to use for grayscale images
     * @param srcImages The image data to use, single for base level
     * @param num The valid number of images to use from the array
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setImages(int mipMapMode,
                          int format,
                          TextureComponent[] srcImages,
                          int num)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        // TODO: validity checking on these values
        width = -1;
        this.format = format;
        this.mipMapMode = mipMapMode;

        if(images == null || images.length < num)
            images = new TextureComponent[num];
        else
        {
            for(int i = numImages; i < num; i++)
                images[i] = null;
        }

        System.arraycopy(srcImages, 0, images, 0, num);
        numImages = num;
        stateChanged = true;
    }

    /**
     * Get the texture type. This returns one of the GL constant types that
     * represent the texture type - 1D, 2D, etc.
     *
     * @return The type constant of this texture
     */
    public int getTextureType()
    {
        return textureType;
    }

    /**
     * Get the format for this texture.
     *
     * @return The format.
     */
    public int getFormat()
    {
        return format;
    }

    /**
     * Get the width of the texture in pixels. If no image is set, this returns
     * -1.
     *
     * @return a number >= -1
     */
    public int getWidth()
    {
        return width;
    }

    /**
     * Set the aniostropic filtering mode.
     *
     * @param mode The new mode.
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setAnisotropicFilterMode(int mode)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        anisotropicMode = mode;
        stateChanged = true;
    }

    /**
     * Get the current aniostropic filtering mode.
     *
     * @return The current mode.
     */
    public int getAnisotropicFilterMode()
    {
        return anisotropicMode;
    }

    /**
     * Set the anisotropic filtering degree.  Values greater
     * then the hardware supports will be clamped.
     *
     * @param degree The filtering degree.  1.0 is the default.
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setAnisotropicFilterDegree(float degree)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        anisotropicDegree = degree;

        // TODO: Need to clamp to hardware max
        stateChanged = true;
    }

    /**
     * Get the current anisotropic filtering degree. The value returned will
     * be the clamped maximum for the hardware support.
     *
     * @return The filtering degree.  1.0 is the default.
     */
    public float getAnisotropicFilterDegree()
    {
        return anisotropicDegree;
    }

    /**
     * Set the magnification filtering mode.
     *
     * @param mode The new mode.
     */
    public void setMagFilter(int mode)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        magFilter = mode;
        stateChanged = true;
    }

    /**
     * Get the magnification filtering mode.
     *
     * @return The current mode.
     */
    public int getMagFilter()
    {
        return magFilter;
    }

    /**
     * Set the magnification filtering mode.
     *
     * @param mode The new mode.
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setMinFilter(int mode)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        minFilter = mode;
        stateChanged = true;
    }

    /**
     * Get the minification filtering mode.
     *
     * @return The current mode.
     */
    public int getMinFilter()
    {
        return minFilter;
    }

    /**
     * Set the boundary handling for the S parameter.
     *
     * @param mode The new mode.
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setBoundaryModeS(int mode)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        boundaryModeS = mode;
        stateChanged = true;
    }

    /**
     * Get the current boundary handling for the S parameter.
     *
     * @return The current mode.
     */
    public int getBoundaryModeS()
    {
        return boundaryModeS;
    }

    /**
     * Cleanup the object now for the given GL context.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void cleanup(GL gl, GLU glu)
    {
        if(textureId != -1)
        {
            int tex_id_tmp[] = { textureId };
            gl.glDeleteTextures(1, tex_id_tmp);
            textureId = -1;
        }
    }
}