/*****************************************************************************
 *                        Web3d.org Copyright (c) 2003
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// Standard imports
import java.nio.ByteBuffer;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * Describes the 2D texture properties of an object.
 *
 * @author Alan Hudson
 * @version $Revision: 1.17 $
 */
public class Texture2D extends Texture
{
    /** The boundary mode S value */
    protected int boundaryModeT;

    /** The height of the main texture. */
    protected int height;

    /**
     * Constructs a texture with default values.
     */
    public Texture2D()
    {
        super(GL.GL_TEXTURE_2D);
        height = -1;
    }

    /**
     * Constructs a texture using just a single image, thus setting it initially
     * as MODE_BASE_LEVEL.
     */
    public Texture2D(int format, TextureComponent2D singleImage)
    {
        super(GL.GL_TEXTURE_2D);

        this.format = format;
        images = new TextureComponent[1];
        images[0] = singleImage;

        width = singleImage.getWidth();
        height = singleImage.getHeight();
        stateChanged = true;
        numImages = 1;
    }

    /**
     * Constructs a 2D texture with the specified images.  The number of images
     * should be 1 for no mip maps, or all mipmap levels should be provided.
     *
     * @param srcImages The image defining this texture.
     * @param num The number of valid images to take from the array
     */
    public Texture2D(int format, TextureComponent2D[] srcImages, int num)
    {
        super(GL.GL_TEXTURE_2D);

        this.format = format;

        if(srcImages != null)
        {
            if(num == 1)
                mipMapMode = MODE_BASE_LEVEL;
            else
                mipMapMode = MODE_MIPMAP;
        }

        images = new TextureComponent[num];
        System.arraycopy(srcImages, 0, images, 0, num);

        width = srcImages[0].getWidth();
        height = srcImages[0].getHeight();
        stateChanged = true;
        numImages = num;
    }

    /**
     * Issue ogl commands needed for this component
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void renderState(GL gl, GLU glu)
    {
        if(numImages == 0)
            return;

        if(!stateChanged)
        {
            if(textureId != -1)
                gl.glBindTexture(textureType, textureId);

            return;
        }

        stateChanged = false;

        // NOTE:
        // Assume, for now, that we need to regenerate the entire texture
        // because the imageComponent has changed. Once sub-image update is
        // implemented, we may want to change this logic
        int[] tex_id_tmp = new int[1];
        gl.glGenTextures(1, tex_id_tmp);
        textureId = tex_id_tmp[0];

        gl.glBindTexture(textureType, textureId);

        gl.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1);
        gl.glTexParameteri(GL.GL_TEXTURE_2D,
                           GL.GL_TEXTURE_WRAP_S,
                           boundaryModeS);

        gl.glTexParameteri(GL.GL_TEXTURE_2D,
                           GL.GL_TEXTURE_WRAP_T,
                           boundaryModeT);
        int mode = 0;
        switch(magFilter)
        {
            case MAGFILTER_FASTEST:
            case MAGFILTER_BASE_LEVEL_POINT:
                mode = GL.GL_NEAREST;
                break;

            case MAGFILTER_NICEST:
            case MAGFILTER_BASE_LEVEL_LINEAR:
                mode = GL.GL_LINEAR;
                break;

            default:
                System.out.println("Unknown mode in MagFilter: " + magFilter);
        }

        gl.glTexParameteri(GL.GL_TEXTURE_2D,
                           GL.GL_TEXTURE_MAG_FILTER,
                           mode);

        switch(minFilter)
        {
            case MINFILTER_FASTEST:
            case MINFILTER_BASE_LEVEL_POINT:
                mode = GL.GL_NEAREST;
                break;

            case MINFILTER_BASE_LEVEL_LINEAR:
                mode = GL.GL_LINEAR;
                break;

            case MINFILTER_MULTI_LEVEL_LINEAR:
                mode = GL.GL_LINEAR_MIPMAP_LINEAR;
                break;

            case MINFILTER_MULTI_LEVEL_POINT:
                mode = GL.GL_NEAREST_MIPMAP_NEAREST;
                break;

            case MINFILTER_NICEST:
                mode = (numImages > 1) ? GL.GL_LINEAR_MIPMAP_LINEAR : GL.GL_LINEAR;
                break;

            default:
                System.out.println("Unknown mode in MinFilter: " + minFilter);
        }

        gl.glTexParameteri(GL.GL_TEXTURE_2D,
                           GL.GL_TEXTURE_MIN_FILTER,
                           mode);

        if(anisotropicMode != ANISOTROPIC_MODE_NONE)
        {
            // float[] val = new float[1];

            //gl.glGetFloatv(GL.GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, val);
            gl.glTexParameterf(GL.GL_TEXTURE_2D,
                               GL.GL_TEXTURE_MAX_ANISOTROPY_EXT,
                               anisotropicDegree);
        }

        int img_count = (mipMapMode == MODE_BASE_LEVEL) ? 1 : images[0].getNumLevels();

        int width = images[0].getWidth();
        int height = ((TextureComponent2D)images[0]).getHeight();

        for(int i = 0; i < img_count; i++)
        {
            ByteBuffer pixels = images[0].getData(i);

            int comp_format = images[0].getFormat(i);
            int int_format = GL.GL_RGB;
            int ext_format = GL.GL_RGB;

            switch(comp_format)
            {
                case TextureComponent.FORMAT_RGB:
                    int_format = GL.GL_RGB;
                    ext_format = GL.GL_RGB;
                    break;

                case TextureComponent.FORMAT_RGBA:
                    int_format = GL.GL_RGBA;
                    ext_format = GL.GL_RGBA;
                    break;

                case TextureComponent.FORMAT_BGR:
                    int_format = GL.GL_BGR;
                    ext_format = GL.GL_BGR;
                    break;

                case TextureComponent.FORMAT_BGRA:
                    int_format = GL.GL_BGRA;
                    ext_format = GL.GL_BGRA;
                    break;


                case TextureComponent.FORMAT_INTENSITY_ALPHA:
                    int_format = GL.GL_LUMINANCE_ALPHA;
                    ext_format = GL.GL_LUMINANCE_ALPHA;
                    break;

                case TextureComponent.FORMAT_SINGLE_COMPONENT:
                    switch(format)
                    {
                        case FORMAT_INTENSITY:
                            int_format = GL.GL_LUMINANCE;
                            ext_format = GL.GL_LUMINANCE;
                            break;

                        case FORMAT_ALPHA:
                            int_format = GL.GL_ALPHA;
                            ext_format = GL.GL_ALPHA;
                    }
                    break;

                default:
            }

            gl.glTexImage2D(GL.GL_TEXTURE_2D,
                            i,
                            int_format,
                            width,
                            height,
                            0,
                            ext_format,
                            GL.GL_UNSIGNED_BYTE,
                            pixels);

            if (width > 1)
                width = width >> 1;
            if (height > 1)
                height = height >> 1;
        }
    }

    /**
     * Restore all openGL state to the given drawable
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void restoreState(GL gl, GLU glu)
    {
        gl.glDisable(textureType);
    }

    /**
     * Set the boundary handling for the T parameter.
     *
     * @param mode The new mode.
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setBoundaryModeT(int mode)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        boundaryModeT = mode;
        stateChanged = true;
    }

    /**
     * Get the current boundary handling for the T parameter.
     *
     * @return The current mode.
     */
    public int getBoundaryModeT()
    {
        return boundaryModeT;
    }

    /**
     * Get the height of the texture in pixels. If no image is set, this returns
     * -1.
     *
     * @return a number >= -1
     */
    public int getHeight()
    {
        return height;
    }
}