/*****************************************************************************
 *                        Web3d.org Copyright (c) 2001
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// Standard imports
import java.awt.image.*;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.IntBuffer;

import java.util.HashMap;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * A component that contains image data to be used in textures.
 *
 * @author Justin Couch, Alan Hudson
 * @version $Revision: 1.3 $
 */
public abstract class TextureComponent extends NodeComponent
{
    /** Specifies the image is in RGB format */
    public static final int FORMAT_RGB = 1;

    /** Specifies the image is in RGBA format */
    public static final int FORMAT_RGBA = 2;

    /** Specifies the image is in Windows BGR format */
    public static final int FORMAT_BGR = 3;

    /** Specifies the image is in Windows BGRA format */
    public static final int FORMAT_BGRA = 4;

    /** Specifies the image is in 2-component Intensity-Alpha format */
    public static final int FORMAT_INTENSITY_ALPHA = 5;

    /** Specifies the image is in 1-component Intensity or Alpha format */
    public static final int FORMAT_SINGLE_COMPONENT = 6;

    /** Specifies the data is in byte format */
    public static final int TYPE_BYTE = DataBuffer.TYPE_BYTE;

    /** Specifies the data is in int format */
    public static final int TYPE_INT = DataBuffer.TYPE_INT;

    /** The width */
    protected int width;

    /** The format */
    protected int format;

    /** The size of the data buffer */
    protected int size;

    /** The type of the data */
    protected int type;

    /** Buffer to hold the data */
    protected ByteBuffer[] data;

    /** Flag describing whether the Y axis should be inverted before use. By default, yes */
    protected boolean invertY;

    /** The number of levels in this component. */
    protected int numLevels;

    /**
     * Constructs an image with default values.
     *
     * @param numLevels The number of mipmap levels to create
     */
    public TextureComponent(int numLevels)
    {
        invertY = true;
        data = new ByteBuffer[numLevels];
        this.numLevels = numLevels;
    }

    /**
     * Get the number of levels in this component.
     *
     * @return The number of levels.
     */
    public int getNumLevels() {
        return numLevels;
    }

    /**
     * Get the width of this image.
     *
     * @return the width.
     */
    public int getWidth()
    {
        return width;
    }

    /**
     * Get the format of this image.
     *
     * @return the format.
     */
    public int getFormat(int level)
    {
        // force a conversion now so that format is set correctly.
        if(data[level] == null)
            data[level] = convertImage(level);

        return format;
    }

    /**
     * Get the current value of the Y-axis inversion flag.
     *
     * @return true if the Y axis should be flipped
     */
    public boolean isYUp()
    {
        return invertY;
    }

    /**
     * Get the underlying data object.  This will be
     * an array of the underlying data type.
     *
     * @param level Which image level needs to be converted
     * @return A reference to the data.
     */
    protected ByteBuffer getData(int level)
    {
        if(data[level] == null)
            data[level] = convertImage(level);

        return data[level];
    }

    /**
     * Clear the storage used for this object.
     *
     * @param level Which image level needs to be converted
     */
    protected void clearData(int level)
    {
        data[level] = null;
    }

    /**
     * Convenience method to convert a buffered image into a NIO array of the
     * corresponding type. Images typically need to be swapped when doing this
     * by the Y axis is in the opposite direction to the one used by OpenGL.
     *
     * @param level Which image level needs to be converted
     * @return an appropriate array type - either IntBuffer or ByteBuffer
     */
    protected abstract ByteBuffer convertImage(int level);
}