/*****************************************************************************
 *                        Web3d.org Copyright (c) 2001
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// External imports
import java.awt.image.*;

import java.nio.ByteBuffer;
import java.nio.FloatBuffer;
import java.nio.ByteOrder;
import java.util.HashMap;

// Local imports

/**
 * Common representation of vertex-based geometry.
 *
 * @author Justin Couch
 * @version $Revision: 1.21 $
 */
public abstract class VertexGeometry extends Geometry
{
    /** 2D Coordinate information is included in the vertex values */
    public static final int COORDINATE_2 = 0x02;

    /** 3D Coordinate information is included in the vertex values */
    public static final int COORDINATE_3 = 0x03;

    /** 4D Coordinate information is included in the vertex values */
    public static final int COORDINATE_4 = 0x04;

    /** Mask to work out what coordinates are used */
    protected static final int COORDINATE_MASK = 0x07;

    /** Mask to clear the coordinate setting */
    protected static final int COORDINATE_CLEAR = 0xFFFFFFF8;

    /** Normal information is included in the vertex values */
    protected static final int NORMALS = 0x08;

    /** Mask to clear the normal setting */
    protected static final int NORMAL_CLEAR = 0xFFFFFFF7;

    /** Mask to work out if texture coordinates are used */
    protected static final int TEXTURE_MASK = 0xF0;

    /** Mask to clear the texture coordinate setting */
    protected static final int TEXTURE_CLEAR = 0xFFFFFF0F;

    /** Single set of texture coordinates are included in the vertex values */
    protected static final int TEXTURE_COORDINATE_SINGLE = 0x10;

    /** Multiple sets of texture coordinates are included in the vertex values */
    protected static final int TEXTURE_COORDINATE_MULTI = 0x20;

    /** Has a valid texture set provided */
    protected static final int TEXTURE_SET_AVAILABLE = 0x0100;

    /** Has a valid texture set provided */
    protected static final int TEXTURE_SET_CLEAR = 0xFFFFF0FF;

    /** Mask to work out if colours are used */
    protected static final int COLOR_MASK = 0xF000;

    /** Mask to clear the texture coordinate setting */
    protected static final int COLOR_CLEAR = 0xFFFF0FFF;

    /** RGB colour values are supplied in the data */
    protected static final int COLOR_3 = 0x1000;

    /** RGBA colour values are supplied in the data */
    protected static final int COLOR_4 = 0x2000;


    // IDs for the texture set provision

    /** 1D texture coordinates are included in the vertex values */
    public static final int TEXTURE_COORDINATE_1 = 1;

    /** 2D texture coordinates are included in the vertex values */
    public static final int TEXTURE_COORDINATE_2 = 2;

    /** 3D texture coordinates are included in the vertex values */
    public static final int TEXTURE_COORDINATE_3 = 3;

    /** 4D texture coordinates are included in the vertex values */
    public static final int TEXTURE_COORDINATE_4 = 4;


    /** The current 2D coordinate list that we work from */
    private float[] working2dCoords;

    /** Working places for a single quad/triangle */
    protected float[] wkPolygon;

    /** A mapping between glContext and displayListID(Integer) */
    protected HashMap displayListMap;

    /** Buffer for holding vertex data */
    protected FloatBuffer vertexBuffer;

    /** Buffer for holding colour data */
    protected FloatBuffer colorBuffer;

    /** Buffer for holding normal data */
    protected FloatBuffer normalBuffer;

    /** Buffer for holding texture coordinate data */
    protected FloatBuffer[] textureBuffer;

    /** Reference to the user array of coordinates used */
    protected float[] coordinates;

    /** Number of valid entries in the coordinate array */
    protected int numCoords;

    /** Reference to the user array of normals used */
    protected float[] normals;

    /** Reference to the user array of textures used, indexed by set */
    protected float[][] textures;

    /** The number of valid texture arrays in the textures variable */
    protected int numTextureArrays;

    /** The texture set map array that describes how to map the arrays */
    protected int[] textureSets;

    /** Flags for the texture type for each array. */
    protected int[] textureTypes;

    /** List of offsets into the texture array
    /** The number of texture sets to use  from the textureSet array */
    protected int numTextureSets;

    /** Reference to the user array of colors used */
    protected float[] colors;

    /** The format of the geometry used */
    protected int vertexFormat;

    /**
     * Flag to say that the display lists must be cleared and regenerated
     * because some state changed
     */
    protected boolean stateChanged;

    /**
     * Constructs an instance with pre-defined values with default values.
     */
    protected VertexGeometry()
    {
        displayListMap = new HashMap(1);
        vertexBuffer = FloatBuffer.allocate(0);
        normalBuffer = FloatBuffer.allocate(0);
        colorBuffer = FloatBuffer.allocate(0);

        textureBuffer = new FloatBuffer[1];
    }

    /**
     * Notification that this object is live now. Overridden to make sure that
     * the live state of the nodes represents the same state as the parent
     * scene graph.
     *
     * @param state true if this should be marked as live now
     */
    protected void setLive(boolean state)
    {
        boolean old_state = alive;

        super.setLive(state);

        if(!old_state && state)
            recomputeBounds();
    }

    //----------------------------------------------------------
    // Local Methods
    //----------------------------------------------------------

    /**
     * Get the current vertex format type - 2D, 3D, or 4D.
     *
     * @return The number of dimensions to the coordinates - 2D, 3D or 4D
     */
    public int getVertexType()
    {
        return (vertexFormat & 0x07);
    }

    /**
     * Get the number of vertices that are valid in the geometry arrays.
     *
     * @return a number >= 0
     * @throws IllegalStateException This call was not made during the update
     *    callback if the node is live
     */
    public int getValidVertexCount()
    {
        return numCoords;
    }

    /**
     * Set the number of vertices to the new number.
     *
     * @param count The new number, must be >= 0
     * @throws IllegalStateException This call was not made during the update
     *    callback if the node is live
     * @throws IllegalArgumentException The number is negative
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setValidVertexCount(int count)
        throws IllegalStateException,
               IllegalArgumentException,
               InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(count < 0)
            throw new IllegalArgumentException("Vertex count is negative");

        stateChanged = true;
        numCoords = count;
    }

    /**
     * Set the vertex array reference to the new array. The number of valid
     * items is taken to be the length of the array divided by three. This
     * replaces the existing vertex list array reference with the new reference.
     *
     * @param type The number of dimensions to the coordinates - 2D, 3D or 4D
     * @param vertices The new array reference to use for vertex information
     * @throws IllegalStateException This call was not made during the update
     *    callback if the node is live
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setVertices(int type, float[] vertices)
        throws IllegalStateException, InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        int num_vert = (vertices == null) ? 0 : vertices.length / type;
        setVertices(type, vertices, num_vert);
    }

    /**
     * Retrieve the vertices that are currently set. The array must be at
     * least as long as the valid vertex count, times 3.
     *
     * @param vertices The array to copy the values into
     * @throws ArrayIndexOutOfBoundsException The provided array is too short
     */
    public void getVertices(float[] vertices)
    {
        int vtx_size = vertexFormat & 0x07;
        System.arraycopy(coordinates, 0, vertices, 0, numCoords * vtx_size);
    }

    /**
     * Set the vertex array reference to the new array. The number of valid
     * items is taken from the second parameter. This replaces the existing
     * vertex list array reference with the new reference.
     *
     * @param type The number of dimensions to the coordinates - 2D, 3D or 4D
     * @param vertices The new array reference to use for vertex information
     * @param numValid The number of valid values to use in the array
     * @throws IllegalStateException This call was not made during the update
     *    callback if the node is live
     * @throws IllegalArgumentException The number is negative
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setVertices(int type, float[] vertices, int numValid)
        throws IllegalStateException,
               IllegalArgumentException,
               InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(numValid < 0)
            throw new IllegalArgumentException("Vertex count is negative");

        stateChanged = true;
        coordinates = vertices;
        numCoords = numValid;

        if(numValid * 3 > vertexBuffer.capacity())
            vertexBuffer = createBuffer(numValid * 3);
        else
            vertexBuffer.clear();

        vertexBuffer.put(vertices, 0, numValid * 3);
        vertexBuffer.rewind();

        if(numCoords == 0)
            vertexFormat &= COORDINATE_CLEAR;
        else
            vertexFormat |= type;
    }

    /**
     * Set the color array reference to the new array. The number of valid
     * items is taken to be the length of the array divided by three. This
     * replaces the existing vertex list array reference with the new reference.
     *
     * @param hasAlpha true if this is 4 component colour, false for 3 component
     * @param colors The new array reference to use for color information
     * @throws IllegalStateException This call was not made during the update
     *    callback if the node is live
     * @throws IllegalArgumentException The length of the colors array is less
     *    than the number of declared vertices
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setColors(boolean hasAlpha, float[] colors)
        throws IllegalStateException,
               IllegalArgumentException,
               InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        int num_valid = 0;

        // check based on format
        if(colors != null)
        {
            if(hasAlpha)
            {
                if(colors.length / 4 < numCoords)
                    throw new IllegalArgumentException("Color array too short " +
                                                   "for 4 component colour");

                vertexFormat |= COLOR_4;
                num_valid = numCoords * 4;
            }
            else
            {
                if(colors.length / 3 < numCoords)
                    throw new IllegalArgumentException("Color array too short " +
                                                   "for 3 component colour");
                vertexFormat |= COLOR_3;
                num_valid = numCoords * 3;
            }
        }
        else
        {
            vertexFormat &= COLOR_CLEAR;
        }

        stateChanged = true;
        this.colors = colors;

        if(num_valid > colorBuffer.capacity())
            colorBuffer = createBuffer(num_valid);
        else
            colorBuffer.clear();

        colorBuffer.put(colors, 0, num_valid);
        colorBuffer.rewind();
    }

    /**
     * Set the normal array reference to the new array. The number of valid
     * items is taken to be the length of the array divided by three. This
     * replaces the existing normal list array reference with the new reference.
     *
     * @param normals The new array reference to use for normal information
     * @throws IllegalStateException This call was not made during the update
     *    callback if the node is live
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setNormals(float[] normals)
        throws IllegalStateException,
               IllegalArgumentException,
               InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if((normals != null) && (normals.length < numCoords * 3))
            throw new IllegalArgumentException("Normal array too short");

        stateChanged = true;
        this.normals = normals;

        if(numCoords * 3 > normalBuffer.capacity())
            normalBuffer = createBuffer(numCoords * 3);
        else
            normalBuffer.clear();

        normalBuffer.put(normals, 0, numCoords * 3);
        normalBuffer.rewind();

        if(normals == null)
            vertexFormat &= NORMAL_CLEAR;
        else
            vertexFormat |= NORMALS;
    }

    /**
     * Set the texture set map to the new mapping. The number of sets defined
     * is the length of the array.
     *
     * @param set The new set to use
     * @throws IllegalStateException This call was not made during the update
     *    callback if the node is live
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setTextureSetMap(int[] set)
        throws IllegalStateException, InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        setTextureSetMap(set, (set == null) ? 0 : set.length);
    }

    /**
     * Set the texture set map to the new mapping. The number of sets defined
     * is the numValid parameter.
     *
     * @param set The new set to use
     * @param numValid The length of the set to use
     * @throws IllegalStateException This call was not made during the update
     *    callback if the node is live
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setTextureSetMap(int[] set, int numValid)
        throws IllegalStateException, InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(set == null || numValid == 0)
        {
            vertexFormat &= TEXTURE_SET_CLEAR;
            numTextureSets = 0;
        }
        else
        {
            numTextureSets = numValid;
            vertexFormat |= TEXTURE_SET_AVAILABLE;
        }

        textureSets = set;
        stateChanged = true;
    }

    /**
     * Set a single texture array reference to the new array. The number of valid
     * items is taken to be the length of the array divided by three. This
     * replaces the existing tex coord list array reference with the new reference.
     *
     * @param type The texture type - 1D, 2D, 3D, 4D.
     * @param textureSet The set to update with these arrays
     * @param texCoords The new array reference to use for vertex information
     * @throws IllegalStateException This call was not made during the update
     *    callback if the node is live
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setTextureCoordinates(int type,
                                      int textureSet,
                                      float[] texCoords)
        throws IllegalStateException,
               IllegalArgumentException,
               InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(textureSet < 0 || textureSet >= numTextureArrays)
            throw new IllegalArgumentException("Invalid texture set specified: " +
                                               textureSet);

        int num_valid = 0;

        if(texCoords != null)
        {
            switch(type)
            {
                case TEXTURE_COORDINATE_1:
                    if(texCoords.length < numCoords)
                        throw new IllegalArgumentException("texCoord array too short " +
                                                       "for 1D texture coordiantes");
                    break;

                case TEXTURE_COORDINATE_2:
                    if(texCoords.length / 2 < numCoords)
                        throw new IllegalArgumentException("texCoord array too short " +
                                                       "for 2D texture coordiantes");
                    break;

                case TEXTURE_COORDINATE_3:
                    if(texCoords.length / 3 < numCoords)
                        throw new IllegalArgumentException("texCoord array too short " +
                                                       "for 3D texture coordiantes");
                    break;

                case TEXTURE_COORDINATE_4:
                    if(texCoords.length / 4 < numCoords)
                        throw new IllegalArgumentException("texCoord array too short " +
                                                       "for 4D texture coordiantes");
                    break;

                default:
                    throw new IllegalArgumentException("Invalid texture type: " + type);
            }
        }

        stateChanged = true;

        if(texCoords.length> textureBuffer[textureSet].capacity())
            textureBuffer[textureSet] = createBuffer(texCoords.length);
        else
            textureBuffer[textureSet].clear();

        textureBuffer[textureSet].put(texCoords, 0, texCoords.length);
    }

    /**
     * Replace all the texture array reference with the new array. The number of
     * valid items is taken to be the length of the array divided by the vertex
     * format defined for this instance. This replaces the existing tex coord
     * list array reference with the new reference.
     *
     * @param types The sets of texture coordinate types that match each array
     * @param texCoords The new array reference to use for vertex information
     * @throws IllegalStateException This call was not made during the update
     *    callback if the node is live
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setTextureCoordinates(int[] types, float[][] texCoords)
        throws IllegalStateException,
               IllegalArgumentException,
               InvalidWriteTimingException
    {
        setTextureCoordinates(types,
                              texCoords,
                              (texCoords == null) ? 0 : texCoords.length);
    }

    /**
     * Replace all the texture array reference to the new array. The number of
     * valid texture coordinates is taken from the numValid parameter. The
     * number of available sets is defined by numSets parameter.
     *
     * @param types The sets of texture coordinate types that match each array
     * @param texCoords The new array reference to use for vertex information
     * @param numSets The number of texture sets that are valid
     * @throws IllegalStateException This call was not made during the update
     *    callback if the node is live
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setTextureCoordinates(int[] types,
                                      float[][] texCoords,
                                      int numSets)
        throws IllegalStateException,
               IllegalArgumentException,
               InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if((texCoords == null) || (numSets == 0))
        {
            vertexFormat &= TEXTURE_CLEAR;

            for(int i = 0; i < numTextureArrays; i++)
                textureBuffer[i].clear();

            numTextureArrays = 0;
        }
        else
        {
            if((numSets < 0) || (texCoords.length < numSets))
                throw new IllegalArgumentException("Invalid set size");

            if((types == null) || (types.length < numSets))
                throw new IllegalArgumentException("Not enough types specified");

            if(textureBuffer.length < numSets)
            {
                FloatBuffer[] tmp = new FloatBuffer[numSets];
                System.arraycopy(textureBuffer, 0, tmp, 0, textureBuffer.length);
                textureBuffer = tmp;
            }

            // check the types for all valid values
            for(int i = 0; i < numSets; i++)
            {
                int num_tex_coords = 0;
                switch(types[i])
                {
                    case TEXTURE_COORDINATE_1:
                        if(texCoords[i].length < numCoords)
                            throw new IllegalArgumentException(
                                "Texture coordinate set " + i + " does not have " +
                                "enough values for a 1D texture. It has " +
                                texCoords[i].length + " but requires " + numCoords);


                        if((textureBuffer[i] == null) ||
                           (textureBuffer[i].capacity() < numCoords))
                            textureBuffer[i] = createBuffer(numCoords);
                        else
                            textureBuffer[i].clear();
                        textureBuffer[i].put(texCoords[i], 0, numCoords);
                        break;

                    case TEXTURE_COORDINATE_2:
                        num_tex_coords = numCoords * 2;

                        if(texCoords[i].length < num_tex_coords)
                            throw new IllegalArgumentException(
                                "Texture coordinate set " + i + " does not have " +
                                "enough values for a 2D texture. It has " +
                                texCoords[i].length + " but requires " + num_tex_coords);

                        if((textureBuffer[i] == null) ||
                           (textureBuffer[i].capacity() < num_tex_coords))
                            textureBuffer[i] = createBuffer(num_tex_coords);
                        else
                            textureBuffer[i].clear();
                        textureBuffer[i].put(texCoords[i], 0, num_tex_coords);
                        break;

                    case TEXTURE_COORDINATE_3:
                        num_tex_coords = numCoords * 3;
                        if(texCoords[i].length < num_tex_coords)
                            throw new IllegalArgumentException(
                                "Texture coordinate set " + i + " does not have " +
                                "enough values for a 3D texture. It has " +
                                texCoords[i].length + " but requires " + num_tex_coords);

                        if((textureBuffer[i] == null) ||
                           (textureBuffer[i].capacity() < num_tex_coords))
                            textureBuffer[i] = createBuffer(num_tex_coords);
                        else
                            textureBuffer[i].clear();
                        textureBuffer[i].put(texCoords[i], 0, num_tex_coords);
                        break;

                    case TEXTURE_COORDINATE_4:
                        num_tex_coords = numCoords * 4;

                        if(texCoords[i].length < num_tex_coords)
                            throw new IllegalArgumentException(
                                "Texture coordinate set " + i + " does not have " +
                                "enough values for a 4D texture. It has " +
                                texCoords[i].length + " but requires " + num_tex_coords);

                        if((textureBuffer[i] == null) ||
                           (textureBuffer[i].capacity() < num_tex_coords))
                            textureBuffer[i] = createBuffer(num_tex_coords);
                        else
                            textureBuffer[i].clear();

                        textureBuffer[i].put(texCoords[i], 0, num_tex_coords);
                        break;

                    default:
                        throw new IllegalArgumentException("Invalid texture type");
                }
            }

            numTextureArrays = numSets;

            if(numTextureSets == 0)
            {
                textureSets = new int[numSets];
                for(int i = 0; i < numSets; i++)
                    textureSets[i] = i;

                numTextureSets = numSets;
            }

            // just have to set some flag to say textures are available.
            vertexFormat |= (numSets > 1) ? TEXTURE_COORDINATE_SINGLE :
                                            TEXTURE_COORDINATE_MULTI;
        }

        textures = texCoords;
        textureTypes = types;

        stateChanged = true;
    }

    /**
     * Mark this node as having dirty bounds due to it's geometry having
     * changed.
     */
    protected void markBoundsDirty()
    {
        for(int i = 0; i < lastParentList; i++)
            parentList[i].markBoundsDirty();
    }

    /**
     * Update this node's bounds and then call the parent to update it's
     * bounds. Used to propogate bounds changes from the leaves of the tree
     * to the root. A node implementation may decide when and where to tell
     * the parent(s)s that updates are ready.
     */
    protected void updateBounds()
    {
        if(numCoords == 0)
            return;

        recomputeBounds();

        for(int i = 0; i < lastParentList; i++)
            parentList[i].updateBounds();
    }

    /**
     * Internal method to recalculate the implicit bounds of this Node. By
     * default the bounds are a point sphere, so derived classes should
     * override this method with something better.
     */
    protected void recomputeBounds()
    {
        if(numCoords == 0)
            return;

        float min_x = coordinates[0];
        float min_y = coordinates[1];
        float min_z = 0;
        float min_w = 0;

        float max_x = coordinates[0];
        float max_y = coordinates[1];
        float max_z = 0;
        float max_w = 0;
        int cnt;

        switch(vertexFormat & COORDINATE_MASK)
        {
            case 2:
                cnt = 2;
                for(int i = 1; i < numCoords; i++)
                {
                    min_x = coordinates[cnt] < min_x ? coordinates[cnt] : min_x;
                    max_x = coordinates[cnt] > max_x ? coordinates[cnt] : max_x;
                    cnt++;

                    min_y = coordinates[cnt] < min_y ? coordinates[cnt] : min_y;
                    max_y = coordinates[cnt] > max_y ? coordinates[cnt] : max_y;
                    cnt++;
                }

                break;

            case 3:
                min_z = coordinates[2];
                max_z = coordinates[2];
                cnt = 3;
                for(int i = 1; i < numCoords; i++)
                {
                    min_x = coordinates[cnt] < min_x ? coordinates[cnt] : min_x;
                    max_x = coordinates[cnt] > max_x ? coordinates[cnt] : max_x;
                    cnt++;

                    min_y = coordinates[cnt] < min_y ? coordinates[cnt] : min_y;
                    max_y = coordinates[cnt] > max_y ? coordinates[cnt] : max_y;
                    cnt++;

                    min_z = coordinates[cnt] < min_z ? coordinates[cnt] : min_z;
                    max_z = coordinates[cnt] > max_z ? coordinates[cnt] : max_z;
                    cnt++;
                }
                break;

            case 4:
                min_z = coordinates[2];
                max_z = coordinates[2];
                min_w = coordinates[3];
                max_w = coordinates[3];
                cnt = 4;

                for(int i = 1; i < numCoords; i++)
                {
                    min_x = coordinates[cnt] < min_x ? coordinates[cnt] : min_x;
                    max_x = coordinates[cnt] > max_x ? coordinates[cnt] : max_x;
                    cnt++;

                    min_y = coordinates[cnt] < min_y ? coordinates[cnt] : min_y;
                    max_y = coordinates[cnt] > max_y ? coordinates[cnt] : max_y;
                    cnt++;

                    min_z = coordinates[cnt] < min_z ? coordinates[cnt] : min_z;
                    max_z = coordinates[cnt] > max_z ? coordinates[cnt] : max_z;
                    cnt++;

// Don't do anything with the 4th coord for now.
//                    min_w = coordinates[cnt] < min_w ? coordinates[cnt] : min_w;
//                    max_w = coordinates[cnt] < max_w ? coordinates[cnt] : max_w;
                    cnt++;
                }
                break;
        }


        if(bounds == null)
            bounds = new BoundingBox();

        BoundingBox bbox = (BoundingBox)bounds;
        bbox.setMinimum(min_x, min_y, min_z);
        bbox.setMaximum(max_x, max_y, max_z);
    }

    /**
     * Check for all intersections against this geometry using a line segment and
     * return the exact distance away of the closest picking point. Default
     * implementation always returns false indicating that nothing was found.
     * Derived classes should override and provide a real implementation.
     *
     * @param start The start point of the segment
     * @param end The end point of the segment
     * @param findAny True if it only has to find a single intersection and can
     *   exit as soon as it finds the first intersection. False if it must find
     *   the closest polygon
     * @param dataOut An array to put the data in for the intersection. Exact
     *   format is described by the flags
     * @param dataOutFlags A set of derived-class specific flags describing what
     *   data should be included in the output array
     * @return True if an intersection was found according to the input request
     * @throws NotPickableException This object has been marked as non pickable,
     *   but you decided to try to call the method anyway
     */
    public boolean pickLineSegment(float[] start,
                                   float[] end,
                                   boolean findAny,
                                   float[] dataOut,
                                   int dataOutFlags)
        throws NotPickableException
    {
        return false;
    }

    /**
     * Check for all intersections against this geometry using a line ray and
     * return the exact distance away of the closest picking point. Default
     * implementation always returns false indicating that nothing was found.
     * Derived classes should override and provide a real implementation.
     *
     * @param origin The start point of the ray
     * @param direction The direction vector of the ray
     * @param findAny True if it only has to find a single intersection and can
     *   exit as soon as it finds the first intersection. False if it must find
     *   the closest polygon
     * @param dataOut An array to put the data in for the intersection. Exact
     *   format is described by the flags
     * @param dataOutFlags A set of derived-class specific flags describing what
     *   data should be included in the output array
     * @return True if an intersection was found according to the input request
     * @throws NotPickableException This object has been marked as non pickable,
     *   but you decided to try to call the method anyway
     */
    public boolean pickLineRay(float[] origin,
                               float[] direction,
                               boolean findAny,
                               float[] dataOut,
                               int dataOutFlags)
        throws NotPickableException
    {
        return false;
    }

    /**
     * Convenience method to allocate a NIO buffer for the vertex handling.
     */
    private FloatBuffer createBuffer(int size)
    {
        // Need to allocate a byte buffer 4 times the size requested because the
        // size is treated as bytes, not number of floats.
        ByteBuffer buf = ByteBuffer.allocateDirect(size * 4);
        buf.order(ByteOrder.nativeOrder());
        FloatBuffer ret_val = buf.asFloatBuffer();

        return ret_val;
    }

    /**
     * Initialize the internal arrays to a given size for the picking.
     *
     * @param numCoords The number of coordinates in the polygon
     */
    protected void initPolygonDetails(int numCoords)
    {
        // assign the working coords to be big enough for a triangle as
        if(working2dCoords == null)
        {
            working2dCoords = new float[numCoords * 2];
            wkPolygon = new float[numCoords * 3];
        }
    }

    /**
     * Private version of the ray - Polygon intersection test that does not
     * do any bounds checking on arrays and assumes everything is correct.
     * Allows fast calls to this method for internal use as well as more
     * expensive calls with checks for the public interfaces.
     * <p>
     * This method does not use wkPoint.
     *
     * @param origin The origin of the ray
     * @param direction The direction of the ray
     * @param length An optional length for to make the ray a segment. If
     *   the value is zero, it is ignored
     * @param dataOut The intersection coordinates and more
     * @return true if there was an intersection, false if not
     */
    protected boolean ray3DTriangleChecked(float[] origin,
                                           float[] direction,
                                           float length,
                                           float[] dataOut)
    {
        int i, j;

        float v0_x = wkPolygon[3] - wkPolygon[0];
        float v0_y = wkPolygon[4] - wkPolygon[1];
        float v0_z = wkPolygon[5] - wkPolygon[2];

        float v1_x = wkPolygon[6] - wkPolygon[3];
        float v1_y = wkPolygon[7] - wkPolygon[4];
        float v1_z = wkPolygon[8] - wkPolygon[5];

        // calculate the normal from the cross product of the two vectors
        float n_x = v0_y * v1_z - v0_z * v1_y;
        float n_y = v0_z * v1_x - v0_x * v1_z;
        float n_z = v0_x * v1_y - v0_y * v1_x;

        // degenerate polygon if the length of the normal is zero
        if(n_x * n_x + n_y * n_y + n_z * n_z == 0)
            return false;

        // normal dot direction
        float n_dot_dir =
            n_x * direction[0] + n_y * direction[1] + n_z * direction[2];

        // ray and plane parallel?
        if(n_dot_dir == 0)
            return false;

        // d dot first point of the polygon
        float d = n_x * wkPolygon[0] +
                  n_y * wkPolygon[1] +
                  n_z * wkPolygon[2];

        float n_dot_o = n_x * origin[0] + n_y * origin[1] + n_z * origin[2];

        // t = (d - N.O) / N.D
        float t = (d - n_dot_o) / n_dot_dir;

        // intersection before the origin
        if(t < 0)
            return false;

        // So we have an intersection with the plane of the polygon and the
        // segment/ray. Using the winding rule to see if inside or outside
        // First store the exact intersection point anyway, regardless of
        // whether this is an intersection or not.
        dataOut[0] = origin[0] + direction[0] * t;
        dataOut[1] = origin[1] + direction[1] * t;
        dataOut[2] = origin[2] + direction[2] * t;

        // Intersection point after the end of the segment?
        if(length != 0)
        {
            float x = origin[0] - dataOut[0];
            float y = origin[1] - dataOut[1];
            float z = origin[2] - dataOut[2];

            if((x * x + y * y + z * z) > (length * length))
                return false;
        }

        // bounds check

        // find the dominant axis to resolve to a 2 axis system
        float abs_nrm_x = (n_x >= 0) ? n_x : -n_x;
        float abs_nrm_y = (n_y >= 0) ? n_y : -n_y;
        float abs_nrm_z = (n_z >= 0) ? n_z : -n_z;

        int dom_axis;

        if(abs_nrm_x > abs_nrm_y)
            dom_axis = 0;
        else
            dom_axis = 1;

        if(dom_axis == 0)
        {
            if(abs_nrm_x < abs_nrm_z)
                dom_axis = 2;
        }
        else if(abs_nrm_y < abs_nrm_z)
        {
            dom_axis = 2;
        }

        // Map all the coordinates to the 2D plane. The u and v coordinates
        // are interleaved as u == even indicies and v = odd indicies

        // Steps 1 & 2 combined
        // 1. For NV vertices [Xn Yn Zn] where n = 0..Nv-1, project polygon
        // vertices [Xn Yn Zn] onto dominant coordinate plane (Un Vn).
        // 2. Translate (U, V) polygon so intersection point is origin from
        // (Un', Vn').
        j = 5;

        switch(dom_axis)
        {
            case 0:
                for(i = 3; --i >= 0; )
                {
                    working2dCoords[j--] = wkPolygon[i * 3 + 2] - dataOut[2];
                    working2dCoords[j--] = wkPolygon[i * 3 + 1] - dataOut[1];
                }
                break;

            case 1:
                for(i = 3; --i >= 0; )
                {
                    working2dCoords[j--] = wkPolygon[i * 3 + 2] - dataOut[2];
                    working2dCoords[j--] = wkPolygon[i * 3]     - dataOut[0];
                }
                break;

            case 2:
                for(i = 3; --i >= 0; )
                {
                    working2dCoords[j--] = wkPolygon[i * 3 + 1] - dataOut[1];
                    working2dCoords[j--] = wkPolygon[i * 3]     - dataOut[0];
                }
                break;
        }

        int sh;  // current sign holder
        int nsh; // next sign holder
        float dist;
        int crossings = 0;

        // Step 4.
        // Set sign holder as f(V' o) ( V' value of 1st vertex of 1st edge)
        if(working2dCoords[1] < 0.0)
            sh = -1;
        else
            sh = 1;

        for(i = 0; i < 3; i++)
        {
            // Step 5.
            // For each edge of polygon (Ua' V a') -> (Ub', Vb') where
            // a = 0..Nv-1 and b = (a + 1) mod Nv

            // b = (a + 1) mod Nv
            j = (i + 1) % 3;

            int i_u = i * 2;           // index of Ua'
            int j_u = j * 2;           // index of Ub'
            int i_v = i * 2 + 1;       // index of Va'
            int j_v = j * 2 + 1;       // index of Vb'

            // Set next sign holder (Nsh) as f(Vb')
            // Nsh = -1 if Vb' < 0
            // Nsh = +1 if Vb' >= 0
            if(working2dCoords[j_v] < 0.0)
                nsh = -1;
            else
                nsh = 1;

            // If Sh <> NSH then if = then edge doesn't cross +U axis so no
            // ray intersection and ignore

            if(sh != nsh)
            {
                // if Ua' > 0 and Ub' > 0 then edge crosses + U' so Nc = Nc + 1
                if((working2dCoords[i_u] > 0.0) && (working2dCoords[j_u] > 0.0))
                {
                    crossings++;
                }
                else if ((working2dCoords[i_u] > 0.0) ||
                         (working2dCoords[j_u] > 0.0))
                {
                    // if either Ua' or U b' > 0 then line might cross +U, so
                    // compute intersection of edge with U' axis
                    dist = working2dCoords[i_u] -
                           (working2dCoords[i_v] *
                            (working2dCoords[j_u] - working2dCoords[i_u])) /
                           (working2dCoords[j_v] - working2dCoords[i_v]);

                    // if intersection point is > 0 then must cross,
                    // so Nc = Nc + 1
                    if(dist > 0)
                        crossings++;
                }

                // Set SH = Nsh and process the next edge
                sh = nsh;
            }
        }

        // Step 6. If Nc odd, point inside else point outside.
        // Note that we have already stored the intersection point way back up
        // the start.
        return ((crossings % 2) == 1);
    }

    /**
     * Private version of the ray - Polygon intersection test that does not
     * do any bounds checking on arrays and assumes everything is correct.
     * Allows fast calls to this method for internal use as well as more
     * expensive calls with checks for the public interfaces.
     * <p>
     * This method does not use wkPoint.
     *
     * @param origin The origin of the ray
     * @param direction The direction of the ray
     * @param length An optional length for to make the ray a segment. If
     *   the value is zero, it is ignored
     * @param dataOut The intersection coordinates and more
     * @return true if there was an intersection, false if not
     */
    protected boolean ray3DQuadChecked(float[] origin,
                                       float[] direction,
                                       float length,
                                       float[] dataOut)
    {
        int i, j;

        float v0_x = wkPolygon[3] - wkPolygon[0];
        float v0_y = wkPolygon[4] - wkPolygon[1];
        float v0_z = wkPolygon[5] - wkPolygon[2];

        float v1_x = wkPolygon[6] - wkPolygon[3];
        float v1_y = wkPolygon[7] - wkPolygon[4];
        float v1_z = wkPolygon[8] - wkPolygon[5];

        // calculate the normal from the cross product of the two vectors
        float n_x = v0_y * v1_z - v0_z * v1_y;
        float n_y = v0_z * v1_x - v0_x * v1_z;
        float n_z = v0_x * v1_y - v0_y * v1_x;

        // degenerate polygon if the length of the normal is zero
        if(n_x * n_x + n_y * n_y + n_z * n_z == 0)
            return false;

        // normal dot direction
        float n_dot_dir =
            n_x * direction[0] + n_y * direction[1] + n_z * direction[2];

        // ray and plane parallel?
        if(n_dot_dir == 0)
            return false;

        // d dot first point of the polygon
        float d = n_x * wkPolygon[0] +
                  n_y * wkPolygon[1] +
                  n_z * wkPolygon[2];

        float n_dot_o = n_x * origin[0] + n_y * origin[1] + n_z * origin[2];

        // t = (d - N.O) / N.D
        float t = (d - n_dot_o) / n_dot_dir;

        // intersection before the origin
        if(t < 0)
            return false;

        // So we have an intersection with the plane of the polygon and the
        // segment/ray. Using the winding rule to see if inside or outside
        // First store the exact intersection point anyway, regardless of
        // whether this is an intersection or not.
        dataOut[0] = origin[0] + direction[0] * t;
        dataOut[1] = origin[1] + direction[1] * t;
        dataOut[2] = origin[2] + direction[2] * t;

        // Intersection point after the end of the segment?
        if(length != 0)
        {
            float x = origin[0] - dataOut[0];
            float y = origin[1] - dataOut[1];
            float z = origin[2] - dataOut[2];

            if((x * x + y * y + z * z) > (length * length))
                return false;
        }

        // bounds check

        // find the dominant axis to resolve to a 2 axis system
        float abs_nrm_x = (n_x >= 0) ? n_x : -n_x;
        float abs_nrm_y = (n_y >= 0) ? n_y : -n_y;
        float abs_nrm_z = (n_z >= 0) ? n_z : -n_z;

        int dom_axis;

        if(abs_nrm_x > abs_nrm_y)
            dom_axis = 0;
        else
            dom_axis = 1;

        if(dom_axis == 0)
        {
            if(abs_nrm_x < abs_nrm_z)
                dom_axis = 2;
        }
        else if(abs_nrm_y < abs_nrm_z)
        {
            dom_axis = 2;
        }

        // Map all the coordinates to the 2D plane. The u and v coordinates
        // are interleaved as u == even indicies and v = odd indicies

        // Steps 1 & 2 combined
        // 1. For NV vertices [Xn Yn Zn] where n = 0..Nv-1, project polygon
        // vertices [Xn Yn Zn] onto dominant coordinate plane (Un Vn).
        // 2. Translate (U, V) polygon so intersection point is origin from
        // (Un', Vn').
        j = 7;   // 2 * numCoords - 1

        switch(dom_axis)
        {
            case 0:
                for(i = 4; --i >= 0; )
                {
                    working2dCoords[j--] = wkPolygon[i * 3 + 2] - dataOut[2];
                    working2dCoords[j--] = wkPolygon[i * 3 + 1] - dataOut[1];
                }
                break;

            case 1:
                for(i = 4; --i >= 0; )
                {
                    working2dCoords[j--] = wkPolygon[i * 3 + 2] - dataOut[2];
                    working2dCoords[j--] = wkPolygon[i * 3]     - dataOut[0];
                }
                break;

            case 2:
                for(i = 4; --i >= 0; )
                {
                    working2dCoords[j--] = wkPolygon[i * 3 + 1] - dataOut[1];
                    working2dCoords[j--] = wkPolygon[i * 3]     - dataOut[0];
                }
                break;
        }

        int sh;  // current sign holder
        int nsh; // next sign holder
        float dist;
        int crossings = 0;

        // Step 4.
        // Set sign holder as f(V' o) ( V' value of 1st vertex of 1st edge)
        if(working2dCoords[1] < 0.0)
            sh = -1;
        else
            sh = 1;

        for(i = 0; i < 4; i++)
        {
            // Step 5.
            // For each edge of polygon (Ua' V a') -> (Ub', Vb') where
            // a = 0..Nv-1 and b = (a + 1) mod Nv

            // b = (a + 1) mod Nv
            j = (i + 1) % 4;

            int i_u = i * 2;           // index of Ua'
            int j_u = j * 2;           // index of Ub'
            int i_v = i * 2 + 1;       // index of Va'
            int j_v = j * 2 + 1;       // index of Vb'

            // Set next sign holder (Nsh) as f(Vb')
            // Nsh = -1 if Vb' < 0
            // Nsh = +1 if Vb' >= 0
            if(working2dCoords[j_v] < 0.0)
                nsh = -1;
            else
                nsh = 1;

            // If Sh <> NSH then if = then edge doesn't cross +U axis so no
            // ray intersection and ignore

            if(sh != nsh)
            {
                // if Ua' > 0 and Ub' > 0 then edge crosses + U' so Nc = Nc + 1
                if((working2dCoords[i_u] > 0.0) && (working2dCoords[j_u] > 0.0))
                {
                    crossings++;
                }
                else if ((working2dCoords[i_u] > 0.0) ||
                         (working2dCoords[j_u] > 0.0))
                {
                    // if either Ua' or U b' > 0 then line might cross +U, so
                    // compute intersection of edge with U' axis
                    dist = working2dCoords[i_u] -
                           (working2dCoords[i_v] *
                            (working2dCoords[j_u] - working2dCoords[i_u])) /
                           (working2dCoords[j_v] - working2dCoords[i_v]);

                    // if intersection point is > 0 then must cross,
                    // so Nc = Nc + 1
                    if(dist > 0)
                        crossings++;
                }

                // Set SH = Nsh and process the next edge
                sh = nsh;
            }
        }

        // Step 6. If Nc odd, point inside else point outside.
        // Note that we have already stored the intersection point way back up
        // the start.
        return ((crossings % 2) == 1);
    }
}
