/*****************************************************************************
 *                        Web3d.org Copyright (c) 2003
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// Standard imports
import java.util.HashMap;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * Node that handles Vertex shader implementation
 * <p>
 *
 * The implementation is relatively simple using just a bind program call followed
 * by the parameter setting. If the application needs greater control than this,
 * such as multi-pass rendering, then derive this class and change the render-loop
 * code to do what is required.
 * <p>
 *
 * Parameters follow the OpenGL model. There are 96 evironment parameters and 96
 * local parameters. Both can be set through this class, though this may change in
 * a future design revision, to make environment parameters into a global setting.
 * <p>
 *
 * Though OpenGL can take the attributes as doubles, this is not supported by this
 * API currently.
 *
 * @author Justin Couch
 * @version $Revision: 1.8 $
 */
public class VertexShader extends ShaderProgram
{

    /** The vertex attributes used by the shader. */
    private float[][] vertexAttributes;

    /** The global environment attributes used by the shader */
    private float[][] envParameters;

    /** The local environment attributes used by the shader */
    private float[][] localParameters;

    /** List of the valid vertex attributes */
    private int[] validAttributes;

    /** The number of valid items in the valid attributes list */
    private int numValidAttributes;

    /** List of the valid environment parameters */
    private int[] validEnvironments;

    /** The number of valid items in the valid environment params list */
    private int numValidEnvironments;

    /** List of the valid local parameters */
    private int[] validLocals;

    /** The number of valid items in the valid local params list */
    private int numValidLocals;

    /**
     * Constructs a Vertex shader with default values. The internal arrays
     * for vertex attributes will be 16 and environment and local parameters
     * to 96.
     */
    public VertexShader()
    {
        this(16, 96);
    }

    /**
     * Create a vertext shader instance with a guaranteed maximum parameter
     * list size. If an attempt is made to read/set past this, an exception
     * will generated. This ID is the maximum parameter index that will be
     * used and the maxium number actually used.
     *
     * @param paramListSize The maximum number of environment and local
     *   parameters that will be used.
     * @param attrListSize The maximum number of vertex attributes that will
     *    be used
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public VertexShader(int attrListSize, int paramListSize)
        throws InvalidWriteTimingException
    {
        vertexAttributes = new float[16][];
        envParameters = new float[paramListSize][];
        localParameters = new float[paramListSize][];

        validAttributes = new int[16];
        validEnvironments = new int[paramListSize];
        validLocals = new int[paramListSize];
    }

    /**
     * Set the environment parameter as a float array.
     *
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setEnvironmentParam(int idx, float[] value)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(envParameters[idx] == null)
            envParameters[idx] = new float[4];

        envParameters[idx][0] = value[0];
        envParameters[idx][1] = value[1];
        envParameters[idx][2] = value[2];
        envParameters[idx][3] = value[3];

        // is it in the list already? If not, add it
        boolean valid = false;
        for(int i = 0; i < numValidEnvironments; i++)
        {
            if(validEnvironments[i] == idx)
            {
                valid = true;
                break;
            }
        }

        if(!valid)
            validEnvironments[numValidEnvironments++] = idx;
    }

    /**
     * Set the local parameter as a float array.
     *
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setLocalParam(int idx, float[] value)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(localParameters[idx] == null)
            localParameters[idx] = new float[4];

        localParameters[idx][0] = value[0];
        localParameters[idx][1] = value[1];
        localParameters[idx][2] = value[2];
        localParameters[idx][3] = value[3];

        // is it in the list already? If not, add it
        boolean valid = false;
        for(int i = 0; i < numValidLocals; i++)
        {
            if(validLocals[i] == idx)
            {
                valid = true;
                break;
            }
        }

        if(!valid)
            validLocals[numValidLocals++] = idx;
    }

    /**
     * Set the vertex attribute as a float array.
     */
    public void setVertexAttribute(int idx, float[] value)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);

        if(vertexAttributes[idx] == null)
            vertexAttributes[idx] = new float[4];

        vertexAttributes[idx][0] = value[0];
        vertexAttributes[idx][1] = value[1];
        vertexAttributes[idx][2] = value[2];
        vertexAttributes[idx][3] = value[3];

        // is it in the list already? If not, add it
        boolean valid = false;
        for(int i = 0; i < numValidAttributes; i++)
        {
            if(validAttributes[i] == idx)
            {
                valid = true;
                break;
            }
        }

        if(!valid)
            validAttributes[numValidAttributes++] = idx;
    }

    /**
     * Set up the rendering state now.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void renderState(GL gl, GLU glu)
    {
        if(programString == null)
            return;

        // do we need to re-generate the program?
        if(programChanged)
        {
            int[] prog_tmp = new int[1];
            gl.glGenProgramsARB(1, prog_tmp);
            programId = prog_tmp[0];

            gl.glBindProgramARB(GL.GL_VERTEX_PROGRAM_ARB, programId);
            gl.glProgramStringARB(GL.GL_VERTEX_PROGRAM_ARB,
                                  GL.GL_PROGRAM_FORMAT_ASCII_ARB,
                                  programString.length(),
                                  programString);
            programChanged = false;
        }

        gl.glBindProgramARB(GL.GL_VERTEX_PROGRAM_ARB, programId);

        for(int i = 0; i < numValidAttributes; i++)
            gl.glVertexAttrib4fvARB(validAttributes[i],
                                    vertexAttributes[validAttributes[i]]);

        for(int i = 0; i < numValidEnvironments; i++)
            gl.glProgramEnvParameter4fvARB(GL.GL_VERTEX_PROGRAM_ARB,
                                           validEnvironments[i],
                                           envParameters[validEnvironments[i]]);

        for(int i = 0; i < numValidLocals; i++)
            gl.glProgramEnvParameter4fvARB(GL.GL_VERTEX_PROGRAM_ARB,
                                           validLocals[i],
                                           localParameters[validLocals[i]]);

        gl.glEnable(GL.GL_VERTEX_PROGRAM_ARB);
    }

    /*
     * This method is called after a node has been rendered.  This method
     * must be re-entrant.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void restoreState(GL gl, GLU glu)
    {
        gl.glDisable(GL.GL_VERTEX_PROGRAM_ARB);
    }
}