/*****************************************************************************
 *                        Web3d.org Copyright (c) 2003
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d;

// Standard imports
import java.util.HashMap;
//import org.web3d.vecmath.Matrix4f;
import javax.vecmath.Matrix4f;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Local imports
// None

/**
 * A viewpoint into the scene.
 * <p>
 *
 * Viewpoints have their own implicit directional light (a headlight) that
 * is controlled separately to the normal in-scenegraph lights. Lights are
 * scoped by default and thus having the viewpoint also have a light following
 * it around is a pain to deal with, you can select on directly on the
 * viewpoint itself. This light is always pointing down the Z axis in the
 * local coordinate system.
 *
 * @author Justin Couch
 * @version $Revision: 1.11 $
 */
public class Viewpoint extends Leaf
{
    private static final float[] LOCATION = { 0, 0, 0, 1 };

    /** Flag indicating the headlight usage state */
    private boolean useHeadlight;

    /**
     * The default constructor where the headlight is implicitly turned off.
     */
    public Viewpoint()
    {
        useHeadlight = false;
    }

    /**
     * Get the current state of the headlight usage.
     *
     * @return true if the headlight is currently active
     */
    public boolean isHeadlightEnabled()
    {
        return useHeadlight;
    }

    /**
     * Turn the local headlight on/off.
     *
     * @param state true to turn the light on
     * @throws InvalidWriteTimingException An attempt was made to write outside
     *   of the NodeUpdateListener callback method
     */
    public void setHeadlightEnabled(boolean state)
        throws InvalidWriteTimingException
    {
        if(updateHandler != null && !updateHandler.isWritePermitted(this))
            throw new InvalidWriteTimingException(WRITE_TIMING_MSG);


        useHeadlight = state;
    }

    /**
     * This method is called to render this node.  All openGL commands needed
     * to render the node should be executed.  Any transformations needed
     * should be added to the transformation stack premultiplied.  This
     * method must be re-entrant as it can be called from multiple
     * places at once.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void render(GL gl, GLU glu)
    {
        if(useHeadlight)
        {
            gl.glEnable(GL.GL_LIGHT0);
            gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, LOCATION);
//            gl.glLightf(GL.GL_LIGHT0, GL.GL_SPOT_CUTOFF, 45);
        }
    }

    /*
     * This method is called after a node has been rendered.  This method
     * must be re-entrant.
     *
     * @param gl The gl context to draw with
     * @param glu The utility context to draw with
     */
    public void postRender(GL gl, GLU glu)
    {
        if(useHeadlight)
            gl.glDisable(gl.GL_LIGHT0);
    }

    /**
     * Mark this node as having dirty bounds due to one of it's children having
     * their bounds changed.
     */
    protected void markBoundsDirty()
    {
        // Ignored as the bounds of a viewpoint are not valid. It is considered
        // a point in space.
    }
}