/*****************************************************************************
 *                        Web3d.org Copyright (c) 2001
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.pipeline;

// External imports
// None

// Local imports
import org.j3d.aviatrix3d.*;

/**
 * The default implementation of the rendering pipeline usable by most
 * applications.
 * <p>
 * This implementation is targeted towards single threaded architectures.
 * After setting the stages, the render command will not return until
 * everything is complete.
 *
 * @author Justin Couch
 * @version $Revision: 1.4 $
 */
public class DefaultRenderPipeline implements RenderPipeline
{
    /** The culling stage to be used */
    private CullStage culler;

    /** The sorting stage to be used */
    private SortStage sorter;

    /** Surface to draw to */
    private DrawableSurface drawable;

    /** The listener to pass cull to sort */
    private CullToSingleSortListener ctsListener;

    /** The listener to pass sort to the drawable listener */
    private SortToSingleDrawListener stdListener;

    /** Root of the scene this pipeline manages */
    private Scene sceneRoot;

    /**
     * Create an instance of the pipeline with nothing registered.
     */
    public DefaultRenderPipeline()
    {
        ctsListener = new CullToSingleSortListener();
        stdListener = new SortToSingleDrawListener();
    }

    /**
     * Construct a pipeline with the sort and cull stages provided.
     *
     * @param ss The sort stage instance to use
     * @param cs The cull stage instance to use
     */
    public DefaultRenderPipeline(CullStage cs, SortStage ss)
    {
        this();

        culler = cs;
        sorter = ss;

        if(cs != null)
            cs.setCulledGeometryReceiver(ctsListener);

        if(ss != null)
        {
            ctsListener.setSorter(ss);
            ss.setSortedGeometryReceiver(stdListener);
        }
    }

    //---------------------------------------------------------------
    // Methods defined by RenderPipeline
    //---------------------------------------------------------------

    /**
     * Register a drawing surface that this pipeline will send its output to.
     * Setting a null value will remove the current drawable surface.
     *
     * @param surface The surface instance to use or replace
     */
    public void setDrawableSurface(DrawableSurface surface)
    {
        stdListener.setSurface(surface);

        drawable = surface;
    }

    /**
     * Get the currently registered drawable surface instance. If none is set,
     * return null.
     *
     * @return The currently set surface instance or null
     */
    public DrawableSurface getDrawableSurface()
    {
        return drawable;
    }

    /**
     * Start the pipeline functioning now. All steps will be called and this
     * method will not return until all are completed and the surface has
     * swapped.
     */
    public void render()
    {
        if(sceneRoot == null)
            return;

        if(culler != null)
        {
            culler.cull(sceneRoot, 0);

            // then draw after the cull is complete
            if(drawable != null)
            {
                drawable.draw();
                drawable.swap();
            }
        }
    }

    /**
     * Set the root of the scene graph to be used by this pipeline. A value of
     * null will remove the scene from being rendered, causing the pipeline to
     * function as a no-op when rendered.
     *
     * @param scene The scene to be rendered
     */
    public void setRenderableScene(Scene scene)
    {
        sceneRoot = scene;
    }

    //---------------------------------------------------------------
    // Local methods
    //---------------------------------------------------------------

    /**
     * Set the sorter instance to be used. If the instance is null, the current
     * sorter is removed.
     *
     * @param ss The sorter instance to use or null
     */
    public void setSorter(SortStage ss)
    {
        ctsListener.setSorter(ss);

        if(sorter != null)
            sorter.setSortedGeometryReceiver(null);

        if(ss != null)
            ss.setSortedGeometryReceiver(stdListener);

        sorter = ss;
    }

    /**
     * Set the cull instance to be used. If the instance is null, the current
     * culler is removed.
     *
     * @param cs The cull instance to use or null
     */
    public void setCuller(CullStage cs)
    {
        if(culler != null)
            culler.setCulledGeometryReceiver(null);

        if(cs != null)
            cs.setCulledGeometryReceiver(ctsListener);

        culler = cs;
    }
}