/*****************************************************************************
 *                        Web3d.org Copyright (c) 2001
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.pipeline;

// External imports
import java.lang.ref.WeakReference;
import java.lang.ref.ReferenceQueue;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLU;

// Local imports
import org.j3d.aviatrix3d.*;

import org.j3d.util.HashSet;

/**
 * Implementation of the RenderPipelineManager that uses a single thread for
 * all of the processing steps.
 * <p>
 *
 * By default the manager does not start of enabled. An explicit enable call
 * will be needed to kick the rendering process off.
 *
 * @author Justin Couch
 * @version $Revision: 1.18 $
 */
public class SingleThreadRenderManager
    implements Runnable, NodeUpdateHandler, RenderPipelineManager
{
    private static final String ACTIVE_RENDERING_MSG =
        "You cannot make this call now because the system is currently " +
        "rendering. To make this call, first disable the rendering. ";

    /** The initial size of the children list */
    private static final int CHANGELIST_START_SIZE = 200;

    /** The increment size of the list if it gets overflowed */
    private static final int CHANGELIST_INCREMENT = 100;

    /** The Scene Graph root */
    private WorldRootGroup root;

    /** The main scene containing all the data we need. */
    private Scene worldScene;

    /** The change requestors for data changed sets */
    private NodeUpdateListener[] dataChangeList;

    /** The change requestors for bounds changed sets */
    private NodeUpdateListener[] boundsChangeList;

    /** The source objects for data changed sets */
    private Object[] dataSourceList;

    /** The source objects for bounds changed sets */
    private Object[] boundsSourceList;

    /** The internal requestors for bounds changed sets */
    private InternalNodeUpdateListener[] boundsInternalList;

    /** The current place to add change requestors */
    private int lastDataChangeItem;

    /** The current place to add change requestors */
    private int lastBoundsChangeItem;

    /** HashSet to determine duplicates in the change List */
    private HashSet dataChangeListenerSet;

    /** HashSet to determine duplicates in the change List */
    private HashSet dataChangeSrcSet;

    /** HashSet to determine duplicates in the change List */
    private HashSet boundsChangeListenerSet;

    /** HashSet to determine duplicates in the change List */
    private HashSet boundsChangeSrcSet;

    /** The renderable surface this manages */
    private DrawableSurface drawable;

    /** Pipeline to manager as part of the drawing process */
    private RenderPipeline pipeline;

    /** Picking handler for managing the picking */
    private PickingHandler pickHandler;

    /** The minimum frame cycle time */
    private int minimumCycleTime;

    /** Has the initial processChangeList happend */
    private boolean initialized;

    /** Current enabled state */
    private boolean enabled;

    /** The thread that contains this runnable */
    private Thread runtimeThread;

    /** Flag that the runtime thread should be terminated at next chance */
    private boolean terminate;

    /** The external observer for keeping updates in check */
    private ApplicationUpdateObserver observer;

    /** Queue for holding deleted textures */
    private ReferenceQueue textureQueue;

    /** Flag for the writable state used by the NodeUpdateHandler query */
    private boolean pickPermitted;

    /** The current object that is having the update callback called */
    private Object writableObject;

    /**
     * Construct a new render manager with no pipelines or renderers
     * registered.
     */
    public SingleThreadRenderManager()
    {
        enabled = false;
        terminate = false;

        pickPermitted = false;

        dataChangeList = new NodeUpdateListener[CHANGELIST_START_SIZE];
        dataSourceList = new Object[CHANGELIST_START_SIZE];
        dataChangeListenerSet = new HashSet(CHANGELIST_START_SIZE);
        dataChangeSrcSet = new HashSet(CHANGELIST_START_SIZE);
        boundsChangeList = new NodeUpdateListener[CHANGELIST_START_SIZE];
        boundsInternalList =
            new InternalNodeUpdateListener[CHANGELIST_START_SIZE];
        boundsSourceList = new Object[CHANGELIST_START_SIZE];
        boundsChangeListenerSet = new HashSet(CHANGELIST_START_SIZE);
        boundsChangeSrcSet = new HashSet(CHANGELIST_START_SIZE);

        pickHandler = new DefaultPickingHandler();
        textureQueue = new ReferenceQueue();

        lastDataChangeItem = 0;
        lastBoundsChangeItem = 0;

        minimumCycleTime = 0;

        root = new WorldRootGroup();
        root.addChild(null);
        root.setUpdateHandler(this);
        root.setLive(true);
    }

    /**
     * Constructs a new DefaultRenderManager for a single channel
     *
     * @param pipe The new pipe instance to be used
     * @param surface The surface to draw to
     */
    public SingleThreadRenderManager(RenderPipeline pipe,
                                     DrawableSurface surface)
    {
        this();

        pipeline = pipe;
        drawable = surface;
    }

    //---------------------------------------------------------------
    // Methods defined by RenderPipelineManager
    //---------------------------------------------------------------

    /**
     * Tell render to start or stop rendering. If currently running, it
     * will wait until all the pipelines have completed their current cycle
     * and will then halt.
     *
     * @param state True if to enable rendering
     */
    public void setEnabled(boolean state)
    {
        if(enabled == state)
            return;

        if(state)
        {
            enabled = true;
            terminate = false;
            runtimeThread = new Thread(this);
            runtimeThread.start();
        }
        else
        {
            enabled = false;
            terminate = true;
            runtimeThread = null;
        }
    }

    /**
     * Get the current render state of the manager.
     *
     * @return true if the manager is currently running
     */
    public boolean isEnabled()
    {
        return enabled;
    }

    /**
     * Force a single render of all pipelines now. Ignores the enabled and
     * cycle time settings to cause a single render at this point in time.
     * If a render is currently in progress, an exception is generated
     *
     * @throws IllegalStateException The system is currently rendering and
     *   should be disabled first.
     */
    public void renderOnce()
        throws IllegalStateException
    {
        if(enabled)
            throw new IllegalStateException(ACTIVE_RENDERING_MSG);

        processChangeList();
        pipeline.render();
    }

    /**
     * Set the minimum duty cycle of the render manager. This is the type in
     * milliseconds that should be the minimum between frames and can be used
     * to throttle the rendering loop to a maximum frame rate should other
     * systems require CPU time.
     *
     * @param cycleTime The minimum time in milliseconds between frames
     */
    public void setMinimumFrameInterval(int cycleTime)
    {
        minimumCycleTime = cycleTime;
    }

    /**
     * Add a pipeline to be rendered to the manager. A duplicate registration
     * or null value is ignored. For this implementation, a call to this method
     * will replace the existing pipeline rather than add an additional one.
     *
     * @param pipe The new pipe instance to be added
     * @throws IllegalStateException The system is currently rendering and
     *   should be disabled first.
     */
    public void addPipeline(RenderPipeline pipe)
        throws IllegalStateException
    {
        if(enabled)
            throw new IllegalStateException(ACTIVE_RENDERING_MSG);

        if(pipeline != null)
            pipeline.setDrawableSurface(null);

        pipeline = pipe;

        if(pipeline != null)
        {
            if(pipeline.getDrawableSurface() == null)
                pipeline.setDrawableSurface(drawable);

            pipeline.setRenderableScene(worldScene);
        }
    }

    /**
     * Remove an already registered pipeline from the manager. A or null value
     * or one that is not currently registered is ignored. For this
     * implementation, a call to this method will remove the existing pipeline
     * if it is the same instance rather than remove an additional one.
     *
     * @param pipe The pipe instance to be removed
     * @throws IllegalStateException The system is currently rendering and
     *   should be disabled first.
     */
    public void removePipeline(RenderPipeline pipe)
        throws IllegalStateException
    {
        if(enabled)
            throw new IllegalStateException(ACTIVE_RENDERING_MSG);

        if(pipe == pipeline)
        {
            pipeline.setDrawableSurface(null);
            pipeline.setRenderableScene(null);
            pipeline = null;
        }
    }

    /**
     * Register an observer that can be used to know when the application is
     * safe to update the scene graph. A value of null will remove the
     * currently set value.
     *
     * @param obs The observer instance to use
     */
    public void setApplicationObserver(ApplicationUpdateObserver obs)
    {
        observer = obs;
    }

    //---------------------------------------------------------------
    // Methods defined by Runnable
    //---------------------------------------------------------------

    /**
     * Run method used to synchronise the internal rendering state and the
     * external state of the canvas. Should never be called directly.
     */
    public void run()
    {
        drawable.setRenderControlThread(runtimeThread);

        while(!terminate)
        {
            // Throttle the apps runtime
            long start_time = System.currentTimeMillis();

            if(!terminate && observer != null)
            {
                pickPermitted = true;
                observer.updateSceneGraph();
                pickPermitted = false;
            }

            if(!terminate)
                processChangeList();

            if(!terminate)
                pipeline.render();

            long now = System.currentTimeMillis();
            long diff = now - start_time;

            if(diff < minimumCycleTime)
            {
                try
                {
                    Thread.sleep(minimumCycleTime - diff);
                }
                catch(InterruptedException ie)
                {
                }
            }

        }

        drawable.setRenderControlThread(null);
    }

    //---------------------------------------------------------------
    // Methods defined by NodeUpdateHandler
    //---------------------------------------------------------------

    /**
     * Check to see if writing to the node is permitted currently.
     *
     * @param src The object that is requesting the check
     * @return true if the end user can write, false if not
     */
    public boolean isWritePermitted(Object src)
    {
        return src == writableObject;
    }

    /**
     * Check to see if picking is permitted currently.
     *
     * @return true if the end user can pick, false if not
     */
    public boolean isPickingPermitted()
    {
        return pickPermitted;
    }


    /**
     * Notify the handler that you have updates to the SG that might alter
     * a node's bounds.
     *
     * @param l The change requestor
     * @param src The object that is passing this listener through.
     * @param intL Internal listener for making callbacks at a later time
     *    to propogate the bounds changes.
     */
    public synchronized void boundsChanged(NodeUpdateListener l,
                                           Object src,
                                           InternalNodeUpdateListener intL)
    {
        // Check for duplicates
        if(boundsChangeSrcSet.contains(src) && boundsChangeListenerSet.contains(l)) {
            return;
        }

        boundsChangeListenerSet.add(l);
        boundsChangeSrcSet.add(src);

        // Add to a single ChangeList for now
        resizeBoundsChangeList();
        boundsChangeList[lastBoundsChangeItem] = l;
        boundsInternalList[lastBoundsChangeItem] = intL;
        boundsSourceList[lastBoundsChangeItem] = src;
        lastBoundsChangeItem++;
    }

    /**
     * Notify the handler that you have updates to the SG that will not
     * alter a node's bounds.
     *
     * @param l The change requestor
     * @param src The object that is passing this listener through.
     */
    public void dataChanged(NodeUpdateListener l, Object src)
    {
        // Check for duplicates
        if(dataChangeSrcSet.contains(src) && dataChangeListenerSet.contains(l)) {
            return;
        }

        dataChangeListenerSet.add(l);
        dataChangeSrcSet.add(src);

        // Add to a single ChangeList for now
        resizeDataChangeList();
        dataChangeList[lastDataChangeItem] = l;
        dataSourceList[lastDataChangeItem] = src;
        lastDataChangeItem++;
    }

    /**
     * The object provided is no longer needed by its parent. Put this on
     * a queue for objects that could potentially be deleted next frame. Used
     * for any scene graph object type that must be explicitly cleaned up
     * during the OpenGL render cycle, such as textures and pBuffers.
     *
     * @param obj The object that is registering itself for deletion
     */
    public void objectRemoved(DeletableSceneGraphObject obj)
    {
        new WeakReference(obj, textureQueue);
    }

    /**
     * Get the picking handler so that we can do some picking operations.
     *
     * @return the current instance of the picking system
     */
    public PickingHandler getPickHandler()
    {
        return pickHandler;
    }

    //---------------------------------------------------------------
    // Misc Internal methods
    //---------------------------------------------------------------

    /**
     * Set the scene for this manager.  null will remove
     * the set scene.
     *
     * @param scene The new scene
     */
    public void setScene(Scene scene)
    {
        Group geom = null;

        if(scene != null)
        {
            geom = scene.getRenderedGeometry();
        }

        worldScene = scene;
        writableObject = root;
        root.setChild(geom, 0);
        writableObject = null;

        if(pipeline != null)
            pipeline.setRenderableScene(scene);
    }

    /**
     * Set the drawable surface that is the output for this render manager.
     *
     * @param surface The surface to use
     */
    public void setDrawableSurface(DrawableSurface surface)
    {
        if(surface != null)
        {
            if(pipeline.getDrawableSurface() == null)
                pipeline.setDrawableSurface(surface);

            surface.setTextureDeletionQueue(textureQueue);
        }
        else
        {
            drawable.setTextureDeletionQueue(null);
        }

        drawable = surface;
    }

    /**
     * Process the changeList now.
     */
    private void processChangeList()
    {
        // Data changed first
        for(int i = 0; i < lastDataChangeItem; i++)
        {
            writableObject = dataSourceList[i];
            dataChangeList[i].updateNodeDataChanges(dataSourceList[i]);
            dataChangeList[i] = null;
            dataSourceList[i] = null;
        }

        writableObject = null;
        lastDataChangeItem = 0;
        dataChangeSrcSet.clear();
        dataChangeListenerSet.clear();

        for(int i = 0; i < lastBoundsChangeItem; i++)
        {
            writableObject = boundsSourceList[i];
            boundsChangeList[i].updateNodeBoundsChanges(boundsSourceList[i]);
            boundsChangeList[i] = null;
            boundsSourceList[i] = null;
        }

        writableObject = null;

        // Now go through and update all the bounds
        for(int i = 0; i < lastBoundsChangeItem; i++)
        {
            boundsInternalList[i].updateBoundsAndNotify();
            boundsInternalList[i] = null;
        }

        lastBoundsChangeItem = 0;
        boundsChangeSrcSet.clear();
        boundsChangeListenerSet.clear();
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    private final void resizeDataChangeList()
    {
        if((lastDataChangeItem + 1) == dataChangeList.length)
        {
            int old_size = dataChangeList.length;
            int new_size = old_size + CHANGELIST_INCREMENT;

            NodeUpdateListener[] tmp_nodes = new NodeUpdateListener[new_size];
            Object[] tmp_src = new Object[new_size];

            System.arraycopy(dataChangeList, 0, tmp_nodes, 0, old_size);
            System.arraycopy(dataSourceList, 0, tmp_src, 0, old_size);

            dataChangeList = tmp_nodes;
            dataSourceList = tmp_src;
        }
    }

    /**
     * Resize the list if needed. Marked as final in order to encourage the
     * compiler to inline the code for faster execution
     */
    private final void resizeBoundsChangeList()
    {
        if((lastBoundsChangeItem + 1) == boundsChangeList.length)
        {
            int old_size = boundsChangeList.length;
            int new_size = old_size + CHANGELIST_INCREMENT;

            NodeUpdateListener[] tmp_nodes = new NodeUpdateListener[new_size];
            InternalNodeUpdateListener[] tmp_int =
                new InternalNodeUpdateListener[new_size];
            Object[] tmp_src = new Object[new_size];

            System.arraycopy(boundsChangeList, 0, tmp_nodes, 0, old_size);
            System.arraycopy(boundsSourceList, 0, tmp_src, 0, old_size);
            System.arraycopy(boundsInternalList, 0, tmp_int, 0, old_size);

            boundsChangeList = tmp_nodes;
            boundsSourceList = tmp_src;
            boundsInternalList = tmp_int;
        }
    }
}