/*****************************************************************************
 *                        Web3d.org Copyright (c) 2001
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.surface;

// External imports
import net.java.games.jogl.*;

import java.lang.ref.Reference;
import java.lang.ref.ReferenceQueue;

import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;

// Local imports
import org.j3d.aviatrix3d.*;
import org.j3d.device.output.elumens.SPI;
import org.j3d.util.IntHashMap;

/**
 * Handles the rendering for a single output device - be it on-screen or off.
 * <p>
 * The code expects that everything is set up before each call of the display()
 * callback. It does not handle any recursive rendering requests as that is
 * assumed to have been sorted out before calling this renderer.
 * <p>
 *
 * The output may be any of the traditional types: pBuffer, screen or memory
 * or any non-traditional type like haptic devices, network streams etc.
 *
 *
 * @author Alan Hudson
 * @version $Revision: 1.4 $
 */
class ElumensRenderingProcessor
    implements GLEventListener
{
    /** The initial size of the children list */
    private static final int LIST_START_SIZE = 400;

    /** Initial depth for the number of transforms above a view,fog or bg */
    private static final int INITIAL_PATH_DEPTH = 64;


    private static final String NOT_LIVE_VP_MSG =
        "The viewpoint given is not part of a live scene graph";

    private static final String SHARED_VP_MSG =
        "The path from the root of the scene graph to the viewpoint contains " +
        "a SharedGroup instance. This is not permitted.";

    private static final String NOT_LIVE_FOG_MSG =
        "The Fog given is not part of a live scene graph";

    private static final String SHARED_FOG_MSG =
        "The path from the root of the scene graph to the fog contains " +
        "a SharedGroup instance. This is not permitted.";

    private static final String NOT_LIVE_BG_MSG =
        "The background given is not part of a live scene graph";

    private static final String SHARED_BG_MSG =
        "The path from the root of the scene graph to the Background contains " +
        "a SharedGroup instance. This is not permitted.";

    /** Generic user data that is assigned to this scene */
    private Object userData;

    /** Pre and Post-render effects processor, if set */
    private RenderEffectsProcessor effectsProcessor;

    /** The current viewpoint instance */
    private Viewpoint currentViewpoint;

    /** Path to the current viewpoint */
    private TransformGroup[] currentViewpointPath;

    /** The current fog instance */
    private Fog currentFog;

    /** Path to the current fog */
    private TransformGroup[] currentFogPath;

    /** The current background instance */
    private Background currentBackground;

    /** Path to the current background */
    private TransformGroup[] currentBackgroundPath;

    /** The current clear colour */
    private float[] clearColor;

    /** Current view environment */
    private ViewEnvironment viewEnvironment;

    /** The default view environment if none set */
    private ViewEnvironment defaultEnvironment;

    /** The current frustum paramaters */
    private double[] currentFrustum;

    /** Flag to say that colour needs to be reset this frame */
    private boolean resetColor;

    /** Local storage of the nodes that need to be rendered */
    private RenderDetails[] renderableList;

    /** Local storage of the nodes that need to be rendered */
    private int[] operationList;

    /** Number of items in the renderable list */
    private int numRenderables;

    /** Flag to say that init has been called on the canvas */
    private boolean initComplete;

    /** The current aspect ratio. Reset each time reshape() is called */
    private double aspectRatio;

    /** Current surface width */
    private int windowWidth;

    /** Current surface height */
    private int windowHeight;

    /** If the view environment or one of it's parameters have changed */
    private boolean viewEnvironmentChanged;

    /** Queue for holding deleted textures */
    private ReferenceQueue textureQueue;

    /** Stack of the available OGL light IDs that could be assigned to lights */
    private Integer[] availableLights;

    /** index in the availableLights list of the last available */
    private int lastLightIdx;

    /** Mapping of the object ID to it's used light ID */
    private IntHashMap lightIdMap;

    private Matrix4f viewMatrix1;
    private Matrix4f viewMatrix2;
    private float[] matrix;
    private Matrix4f vworldMatrix;

    /** The Java wrapper to the native SPI library */
    private SPI spi;

    /** Should we reinitialize SPI */
    private boolean initSpi;

    /** The new number of channels at next SPI initialize */
    private int newNumChannels;

    /** The number of channels to render each frame. */
    private int numChannels;

    private int spiFormat = SPI.SPI_PF_3_CHAN | SPI.SPI_PF_BACKBUFFER;
    private int wallCnt;
    private float near;
    private float far;


    /**
     * Construct handler for rendering objects to the main screen.
     */
    ElumensRenderingProcessor()
    {
        clearColor = new float[4];
        viewMatrix1 = new Matrix4f();
        viewMatrix2 = new Matrix4f();
        matrix = new float[16];
        vworldMatrix = new Matrix4f();

        defaultEnvironment = new ViewEnvironment();
        viewEnvironment = defaultEnvironment;

        resetColor = false;
        initComplete = false;

        numRenderables = 0;

        currentViewpointPath = new TransformGroup[INITIAL_PATH_DEPTH];
        currentFogPath = new TransformGroup[INITIAL_PATH_DEPTH];
        currentBackgroundPath = new TransformGroup[INITIAL_PATH_DEPTH];

        lightIdMap = new IntHashMap();

        // GL_LIGHT0 is assumed to be used by the current viewpoint
        availableLights = new Integer[7];
        availableLights[0] = new Integer(GL.GL_LIGHT1);
        availableLights[1] = new Integer(GL.GL_LIGHT2);
        availableLights[2] = new Integer(GL.GL_LIGHT3);
        availableLights[3] = new Integer(GL.GL_LIGHT4);
        availableLights[4] = new Integer(GL.GL_LIGHT5);
        availableLights[5] = new Integer(GL.GL_LIGHT6);
        availableLights[6] = new Integer(GL.GL_LIGHT7);

        lastLightIdx = 0;

        spi = new SPI();
        newNumChannels = 3;
        spiFormat = SPI.SPI_PF_3_CHAN | SPI.SPI_PF_BACKBUFFER;
        wallCnt = 0;
        initSpi = true;
    }

    //---------------------------------------------------------------
    // Methods defined by GLEventListener
    //---------------------------------------------------------------

    /**
     * Called by the drawable immediately after the OpenGL context is
     * initialized; the GLContext has already been made current when
     * this method is called.
     *
     * @param drawable The display context to render to
     */
    public void init(GLDrawable drawable)
    {
//        drawable.setGL(new TraceGL(drawable.getGL(), System.out));

        GL gl = drawable.getGL();
        GLU glu = drawable.getGLU();

        gl.glClearColor(clearColor[0],
                        clearColor[1],
                        clearColor[2],
                        clearColor[3]);

        gl.glEnable(GL.GL_DEPTH_TEST);
        gl.glEnable(GL.GL_BLEND);
        gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);

        gl.glHint(GL.GL_PERSPECTIVE_CORRECTION_HINT, GL.GL_NICEST);
        gl.glCullFace(GL.GL_BACK);

        updateProjectionMatrix(gl, glu);

        spi = new SPI();

        initializeSPI();
    }

    /**
     * Called by the drawable when the surface resizes itself. Used to
     * reset the viewport dimensions.
     *
     * @param drawable The display context to render to
     */
    public void reshape(GLDrawable drawable,
                        int x,
                        int y,
                        int width,
                        int height)
    {
        windowWidth = width;
        windowHeight = height;

        aspectRatio = (float)width / (float)height;
        viewEnvironment.setAspectRatio(aspectRatio);
        GL gl = drawable.getGL();
        GLU glu = drawable.getGLU();

        gl.glViewport(0, 0, width, height);

        updateProjectionMatrix(gl, glu);

        int w;
        int h;

        if (windowWidth > 1024)
            w = 1024;
        else
            w = windowWidth;

        if (windowHeight > 1024)
            h = 1024;
        else
            h = windowHeight;

        spi.setChanSize(SPI.SPI_3C_LEFT, w, h);
        spi.setChanSize(SPI.SPI_3C_RIGHT, w, h);
        spi.setChanSize(SPI.SPI_3C_TOP, w, h);

        initComplete = true;
    }

    /**
     * Called by the drawable when the display mode or the display device
     * associated with the GLDrawable has changed.
     *
     * @param drawable The display context to render to
     */
    public void displayChanged(GLDrawable drawable,
                               boolean modeChanged,
                               boolean deviceChanged)
    {
//System.out.println("GL display change");

// Need a way to clean up the old buffer here.
//        if(outputTexture != null)
//        {
//            outputTexture.unregisterBuffer(old_gl);
//            outputTexture.registerBuffer(gl, pBuffer);
//        }
    }

    /**
     * Called by the drawable to perform rendering by the client.
     *
     * @param drawable The display context to render to
     */
    public void display(GLDrawable drawable)
    {
        int token;

        GL gl = drawable.getGL();
        GLU glu = drawable.getGLU();

        if (initSpi) {
            initializeSPI();
        }

        renderViewpoint(gl, glu);

        spi.begin();

        for(int wallCnt=0; wallCnt < numChannels; wallCnt++)
        {

            token = getSpiToken(wallCnt);

            spi.preRender(token);

            if(effectsProcessor != null)
                effectsProcessor.preDraw(drawable, userData);


            if(currentBackground != null)
            {
                if(!(currentBackground instanceof ColorBackground))
                    gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);

//                renderViewpoint(gl, glu);

    //            gl.glLoadIdentity();
                currentBackground.render(gl, glu);
                currentBackground.postRender(gl, glu);
            }
            else
            {
                if(resetColor)
                {
                    gl.glClearColor(clearColor[0],
                                    clearColor[1],
                                    clearColor[2],
                                    clearColor[3]);
                }

                gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);

//                renderViewpoint(gl, glu);

    //            gl.glLoadIdentity();
            }

            // Clear the texture queue first so that we have the resources around
            // for rendering the next stage.
            Object old_tex;

            while((old_tex = textureQueue.poll()) != null)
            {
                Reference ref = (Reference)old_tex;
                DeletableSceneGraphObject tex = (DeletableSceneGraphObject)ref.get();
                tex.cleanup(gl, glu);
            }

            // transpose temporarily
            float[] mat = new float[16];

            for(int i = 0; i < numRenderables; i++)
            {
                switch(operationList[i])
                {
                    case RenderOp.START_RENDER:
                        // load the matrix to render
                        gl.glPushMatrix();
                        gl.glMultMatrixf(renderableList[i].transform);

                        renderableList[i].renderable.render(gl, glu);
                        break;

                    case RenderOp.STOP_RENDER:
                        renderableList[i].renderable.postRender(gl, glu);
                        gl.glPopMatrix();
                        break;

                    case RenderOp.START_LIGHT:
                            // Get the next available light ID
                        Integer l_id = availableLights[lastLightIdx++];
                        lightIdMap.put(renderableList[i].id, l_id);

                        int gl_id = l_id.intValue();
                        Light light = (Light)renderableList[i].renderable;

                        // load the matrix to render
                        gl.glPushMatrix();
                        gl.glMultMatrixf(renderableList[i].transform);
                        light.render(gl, glu, gl_id);
                        gl.glPopMatrix();
                        break;

                    case RenderOp.STOP_LIGHT:
                        l_id = (Integer)lightIdMap.get(renderableList[i].id);
                        lightIdMap.remove(renderableList[i].id);

                        gl_id = l_id.intValue();

                        light = (Light)renderableList[i].renderable;
                        light.postRender(gl, glu, gl_id);
                        availableLights[lastLightIdx--] = l_id;
                        break;
                }
            }

            if(effectsProcessor != null)
                effectsProcessor.postDraw(drawable, userData);

            gl.glFlush();

            spi.postRender(token);
        }


        if(currentViewpoint != null)
            currentViewpoint.postRender(gl, glu);

        spi.end();
        gl.glClearColor( 0.0f, 0.0f, 0.4f, 1.0f);
        gl.glClear (gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT);

        switch(numChannels)
        {
            case 1:
                spi.flush(SPI.SPI_1C_FRONT);
                break;
            case 2:
                spi.flush(SPI.SPI_ALL_2_CHAN);
                break;
            case 3:
                spi.flush(SPI.SPI_ALL_3_CHAN);
                break;
            case 4:
                spi.flush(SPI.SPI_ALL_4_CHAN);
                break;
        }
    }

    /**
     * Set the number of channels to display.  This will
     * happen the next render loop.
     */
    public void setNumberOfChannels(int num)
    {
        newNumChannels = num;
        initSpi = true;
    }

    //---------------------------------------------------------------
    // Local Methods
    //---------------------------------------------------------------

    /**
     * Set the queue that holds references to the textures that are possibly
     * ready for deletion. Poll this queue once per frame and delete any
     * items found on the queue. A value of null will clear the queue.
     *
     * @param queue The instance of the queue to use or null
     */
    public void setTextureDeletionQueue(ReferenceQueue queue)
    {
        textureQueue = queue;
    }

    /**
     * Convenience method to assign up all the environment data. Calls the
     * setActiveX() methods in this class.
     *
     * @throws IllegalArgumentException The path contains a SharedGroup or
     *    the node is not live
     */
    public void setEnvironmentData(RenderEnvironmentData data)
        throws IllegalArgumentException
    {
        setActiveView(data.viewpoint);
        setActiveBackground(data.background);
        setActiveFog(data.fog);
        setViewEnvironment(data.viewEnvironment);
        viewEnvironmentChanged=true;

        setRenderEffects(data.effectsProcessor, data.userData);
        currentFrustum = data.viewFrustum;
    }

    /**
     * Set the viewpoint path that should be applied to the current surface.
     * The output drawn will be a combination of this information and that
     * of the view environment.
     *
     * @param vp The instance of the active viewpoint to use
     * @throws IllegalArgumentException The path contains a SharedGroup or
     *    the node is not live
     */
    public void setActiveView(Viewpoint vp)
        throws IllegalArgumentException
    {
        if(vp == null)
        {

            currentViewpoint = vp;
            return;
        }

        if(!vp.isLive())
            throw new IllegalArgumentException(NOT_LIVE_VP_MSG);

        // walk back up the path to find the current root of
        // the scene graph and check that there is no SharedGroup
        // instances along the way.
        Node parent = vp.getParent();
        int path_depth = 0;

        while(parent != null)
        {
            if(parent instanceof SharedGroup)
                throw new IllegalArgumentException(SHARED_VP_MSG);

            parent = parent.getParent();
            path_depth++;
        }

        currentViewpoint = vp;

        if(currentViewpointPath.length < path_depth)
            currentViewpointPath = new TransformGroup[path_depth];
    }

    /**
     * Set the fog that should be applied to the current surface.
     * The output drawn will be a combination of this information and that
     * of the view environment.
     *
     * @param fog The instance of the active fog node
     * @throws IllegalArgumentException The path contains a SharedGroup or
     *    the node is not live
     */
    public void setActiveFog(Fog fog)
        throws IllegalArgumentException
    {
        if(fog == null)
        {
            currentFog = fog;
            return;
        }

        if(!fog.isLive())
            throw new IllegalArgumentException(NOT_LIVE_FOG_MSG);

        // walk back up the path to find the current root of
        // the scene graph and check that there is no SharedGroup
        // instances along the way.
        Node parent = fog.getParent();
        int path_depth = 0;

        while(parent != null)
        {
            if(parent instanceof SharedGroup)
                throw new IllegalArgumentException(SHARED_FOG_MSG);

            parent = parent.getParent();
            path_depth++;
        }

        currentFog = fog;

        if(currentFogPath.length < path_depth)
            currentFogPath = new TransformGroup[path_depth];
    }

    /**
     * Set the background path that should be applied to the current surface.
     * The output drawn will be a combination of this information and that
     * of the view environment.
     *
     * @param bg The instance of the active background
     * @throws IllegalArgumentException The path contains a SharedGroup or
     *    the node is not live
     */
    public void setActiveBackground(Background bg)
        throws IllegalArgumentException
    {
        if(bg == null)
        {
            currentBackground = bg;
            return;
        }

        if(!bg.isLive())
            throw new IllegalArgumentException(NOT_LIVE_BG_MSG);

        // walk back up the path to find the current root of
        // the scene graph and check that there is no SharedGroup
        // instances along the way.
        Node parent = bg.getParent();
        int path_depth = 0;

        while(parent != null)
        {
            if(parent instanceof SharedGroup)
                throw new IllegalArgumentException(SHARED_BG_MSG);

            parent = parent.getParent();
            path_depth++;
        }

        currentBackground = bg;

        if(currentBackgroundPath.length < path_depth)
            currentBackgroundPath = new TransformGroup[path_depth];
    }

    /**
     * Set the view environment that is used to render this surface.
     *
     * @param env The environment instance to use for the render setup
     */
    public void setViewEnvironment(ViewEnvironment env)
    {
        viewEnvironment = env;

        if(env == null)
            viewEnvironment = defaultEnvironment;
    }

    /**
     * Set the rendering effects processor and the associated user data.  The
     * data object will be passed back to the client code through the callback
     * method when the pre/post render calls are done.
     */
    public void setRenderEffects(RenderEffectsProcessor effects, Object data)
    {
        effectsProcessor = effects;
        userData = data;
    }

    /**
     * Set the background colour that this surface should be cleared to before
     * the drawing step. Colours range from 0 to 1 in the normal manner.
     *
     * @param r The red component of the background clear colour
     * @param g The green component of the background clear colour
     * @param b The blue component of the background clear colour
     * @param a The alpha component of the background clear colour
     */
    public void setClearColor(float r, float g, float b, float a)
    {
        clearColor[0] = r;
        clearColor[1] = g;
        clearColor[2] = b;
        clearColor[3] = a;

        resetColor = true;
    }

    /**
     * Update the list of items to be rendered to the current list. Draw them
     * at the next oppourtunity.
     *
     * @param nodes The list of nodes in sorted order
     * @param renderOps Operation to perform on each node
     * @param numValid The number of valid items in the array
     */
    public void setDrawableObjects(RenderDetails[] nodes,
                                   int[] renderOps,
                                   int numValid)
    {
        renderableList = nodes;
        operationList = renderOps;

        numRenderables = numValid;
    }

    /**
     * Get the surface to VWorld transformation matrix.
     *
     * @param The matrix.  It must be preallocated.
     */
    public void getSurfaceToVWorld(Matrix4f matrix) {
        matrix.m00 = vworldMatrix.m00;
        matrix.m01 = vworldMatrix.m01;
        matrix.m02 = vworldMatrix.m02;
        matrix.m03 = vworldMatrix.m03;
        matrix.m10 = vworldMatrix.m10;
        matrix.m11 = vworldMatrix.m11;
        matrix.m12 = vworldMatrix.m12;
        matrix.m13 = vworldMatrix.m13;
        matrix.m20 = vworldMatrix.m20;
        matrix.m21 = vworldMatrix.m21;
        matrix.m22 = vworldMatrix.m22;
        matrix.m23 = vworldMatrix.m23;
        matrix.m30 = vworldMatrix.m30;
        matrix.m31 = vworldMatrix.m31;
        matrix.m32 = vworldMatrix.m32;
        matrix.m33 = vworldMatrix.m33;
    }

    /**
     * Convert a pixel location to surface coordinates.
     *
     * @param x The X coordinate
     * @param y The Y coordinate
     * @param position The converted position.  It must be preallocated.
     */
    public void getPixelLocationInSurface(float x, float y, Point3f position) {
        if (currentFrustum == null) {
            position.x = 0;
            position.y = 0;
            position.z = 0;
        } else {
            position.x = (float) ((currentFrustum[1] - currentFrustum[0]) * (x / windowWidth - 0.5f));
            position.y = (float) ((currentFrustum[3] - currentFrustum[2]) * ((windowHeight - y) / windowHeight - 0.5f));
            position.z = (float) -currentFrustum[4];
        }
    }

    /**
     * Get the Center Eye position in surface coordinates.
     *
     * @param position The current eye position.  It must be preallocated.
     */
    public void getCenterEyeInSurface(Point3f position) {
        position.x = 0;
        position.y = 0;
        position.z = 0;
    }

    /**
     * Render the viewpoint setup
     */
    private void renderViewpoint(GL gl, GLU glu)
    {
        // Will want to check here for the VP being non-live.
        if(currentViewpoint == null)
            return;

        if(viewEnvironmentChanged)
            updateProjectionMatrix(gl, glu);

        gl.glLoadIdentity();

        // start at the VP and work our way upwards.
        Node parent = currentViewpoint.getParent();
        int last_tx = 0;
        while(parent != null)
        {
//System.out.println("parent: " + parent);
            if(parent instanceof TransformGroup)
                currentViewpointPath[last_tx++] = (TransformGroup)parent;

            parent = parent.getParent();
        }

        viewMatrix2.setIdentity();

        if(last_tx != 0)
        {
            for(int i = last_tx - 1; i >= 0; i--)
            {
                currentViewpointPath[i].getTransform(viewMatrix1);
                viewMatrix2.mul(viewMatrix1);
            }
        }
//System.out.println("viewMatrix2: \n" + viewMatrix2);

        currentViewpoint.render(gl, glu);

        viewMatrix2.invert();

        // Transpose in place
        matrix[0] = viewMatrix2.m00;
        matrix[1] = viewMatrix2.m10;
        matrix[2] = viewMatrix2.m20;
        matrix[3] = viewMatrix2.m30;
        matrix[4] = viewMatrix2.m01;
        matrix[5] = viewMatrix2.m11;
        matrix[6] = viewMatrix2.m21;
        matrix[7] = viewMatrix2.m31;
        matrix[8] = viewMatrix2.m02;
        matrix[9] = viewMatrix2.m12;
        matrix[10] = viewMatrix2.m22;
        matrix[11] = viewMatrix2.m32;
        matrix[12] = viewMatrix2.m03;
        matrix[13] = viewMatrix2.m13;
        matrix[14] = viewMatrix2.m23;
        matrix[15] = viewMatrix2.m33;

        gl.glMultMatrixf(matrix);

    }

    /**
     * Update the projection matrix.
     */
    private void updateProjectionMatrix(GL gl, GLU glu)
    {
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glLoadIdentity();

        double near_clip = viewEnvironment.getNearClipDistance();
        double far_clip = viewEnvironment.getFarClipDistance();

        switch(viewEnvironment.getProjectionType())
        {
            case ViewEnvironment.PERSPECTIVE_PROJECTION:
                if (currentFrustum != null) {
                    gl.glFrustum(currentFrustum[0],currentFrustum[1],
                                 currentFrustum[2],currentFrustum[3],
                                 currentFrustum[4],currentFrustum[5]);
                } else {
                    double fov = viewEnvironment.getFieldOfView();

                    glu.gluPerspective(fov, aspectRatio, near_clip, far_clip);
                }

                break;

            case ViewEnvironment.ORTHOGRAPHIC_PROJECTION:
                gl.glOrtho(0,
                           windowWidth,
                           0,
                           windowHeight,
                           near_clip,
                           far_clip);
                break;

            default:
                System.out.println("unknown projection type");
        }

        viewEnvironmentChanged = false;

        gl.glMatrixMode(GL.GL_MODELVIEW);
    }

    /**
     * Get the spiToken based on the current wallCnt and number of channels.
     *
     * @return The token
     */
    private int getSpiToken(int wallCnt)
    {
        switch(numChannels)
        {
            case 1:
                return SPI.SPI_1C_FRONT;
            case 2:
                switch(wallCnt)
                {
                    case 0: return SPI.SPI_2C_LEFT;
                    case 1: return SPI.SPI_2C_RIGHT;
                    default: System.out.println("Unknown wall: " + wallCnt);
                }
            case 3:
                switch(wallCnt)
                {
                    case 0: return SPI.SPI_3C_LEFT;
                    case 1: return SPI.SPI_3C_RIGHT;
                    case 2: return SPI.SPI_3C_TOP;
                    default: System.out.println("Unknown wall: " + wallCnt);
                }
            case 4:
                switch(wallCnt)
                {
                    case 0: return SPI.SPI_4C_LEFT;
                    case 1: return SPI.SPI_4C_RIGHT;
                    case 2: return SPI.SPI_4C_TOP;
                    case 3: return SPI.SPI_4C_BOTTOM;
                    default: System.out.println("Unknown wall: " + wallCnt);
                }
            default:
                System.out.println("Unsupported number of channels in: " + this);
                return 0;
        }
    }

    /**
     * Get the SPI initialization format based on the
     * number of channels.
     *
     * @return the format
     */
    private int getSpiFormat() {
        int bufferType = SPI.SPI_PF_BACKBUFFER;

        switch(numChannels)
        {
            case 1:
                return (SPI.SPI_PF_1_CHAN | bufferType);
            case 2:
                return (SPI.SPI_PF_2_CHAN | bufferType);
            case 3:
                return (SPI.SPI_PF_3_CHAN | bufferType);
            case 4:
                return (SPI.SPI_PF_4_CHAN | bufferType);
            default:
                System.out.println("Unsupported number of channels in: " + this);
                return 0;
        }
    }

    /**
     * Initialize the SPI library
     */
    private void initializeSPI() {
        numChannels = newNumChannels;
        spiFormat = getSpiFormat();
System.out.println("Initialize spi  numChannels: " + numChannels + " format: " + spiFormat);
        spi.initialize(spiFormat, numChannels);
        near = 0.1f;
        far = 3000;

        spi.setNearFar(near,far);

        if (initComplete) {
            int w;
            int h;

            if (windowWidth > 1024)
                w = 1024;
            else
                w = windowWidth;

            if (windowHeight > 1024)
                h = 1024;
            else
                h = windowHeight;

            switch(numChannels)
            {
                case 1:
                    spi.setChanSize(SPI.SPI_1C_FRONT, w, h);
                    break;
                case 2:
                    spi.setChanSize(SPI.SPI_ALL_2_CHAN, w, h);
                    break;
                case 3:
                    spi.setChanSize(SPI.SPI_ALL_3_CHAN, w, h);
                    break;
                case 4:
                    spi.setChanSize(SPI.SPI_ALL_4_CHAN, w, h);
                    break;
            }
        }

        initSpi = false;
    }
}